#!/usr/bin/env python3
from __future__ import annotations

import asyncio
import contextlib
import json
import logging
import random
import ssl
import time
from typing import Any

import aiofiles
import aiomqtt

from qolsys_controller.mqtt_command import (
    MQTTCommand,
    MQTTCommand_Panel,
    MQTTCommand_ZWave,
)

from .enum import PartitionAlarmState, PartitionArmingType, PartitionSystemStatus
from .enum_zwave import ThermostatFanMode, ThermostatMode, ZwaveCommand, ZwaveDeviceClass
from .errors import QolsysMqttError, QolsysSslError, QolsysUserCodeError
from .mdns import QolsysMDNS
from .mqtt_command_queue import QolsysMqttCommandQueue
from .observable import QolsysObservable
from .panel import QolsysPanel
from .pki import QolsysPKI
from .settings import QolsysSettings
from .state import QolsysState
from .task_manager import QolsysTaskManager
from .utils_mqtt import generate_random_mac

LOGGER = logging.getLogger(__name__)


class QolsysController:
    def __init__(self) -> None:
        # QolsysController
        self._state = QolsysState(self)
        self._settings = QolsysSettings(self)
        self._panel = QolsysPanel(self)

        self.connected = False
        self.connected_observer = QolsysObservable()

        # PKI
        self._pki = QolsysPKI(settings=self.settings)

        # Plugin
        self.certificate_exchange_server: asyncio.Server | None = None
        self._task_manager = QolsysTaskManager()
        self._mqtt_command_queue = QolsysMqttCommandQueue()
        self._zone_id: str = "1"

        # MQTT Client
        self.aiomqtt: aiomqtt.Client | None = None
        self._mqtt_task_config_label: str = "mqtt_task_config"
        self._mqtt_task_listen_label: str = "mqtt_task_listen"
        self._mqtt_task_connect_label: str = "mqtt_task_connect"
        self._mqtt_task_ping_label: str = "mqtt_task_ping"

    @property
    def state(self) -> QolsysState:
        return self._state

    @property
    def panel(self) -> QolsysPanel:
        return self._panel

    @property
    def settings(self) -> QolsysSettings:
        return self._settings

    @property
    def mqtt_command_queue(self) -> QolsysMqttCommandQueue:
        return self._mqtt_command_queue

    def is_paired(self) -> bool:
        return (
            self._pki.id != ""
            and self._pki.check_key_file()
            and self._pki.check_cer_file()
            and self._pki.check_qolsys_cer_file()
            and self._pki.check_secure_file()
            and self.settings.check_panel_ip()
            and self.settings.check_plugin_ip()
        )

    async def config(self, start_pairing: bool) -> Any:
        return await self._task_manager.run(self.config_task(start_pairing), self._mqtt_task_config_label)

    async def config_task(self, start_pairing: bool) -> bool:
        LOGGER.debug("Configuring Plugin")

        # Check and created config_directory
        if not self.settings.check_config_directory(create=start_pairing):
            return False

        # Read user file for access code
        loop = asyncio.get_running_loop()
        if not loop.run_in_executor(None, self.panel.read_users_file):
            return False

        # Config PKI
        if self.settings.auto_discover_pki:
            if self._pki.auto_discover_pki():
                self.settings.random_mac = self._pki.formatted_id()
        else:
            self._pki.set_id(self.settings.random_mac)

        # Set mqtt_remote_client_id
        self.settings.mqtt_remote_client_id = "qolsys-controller-" + self._pki.formatted_id()
        LOGGER.debug("Using MQTT remoteClientID: %s", self.settings.mqtt_remote_client_id)

        # Check if plugin is paired
        if self.is_paired():
            LOGGER.debug("Panel is Paired")

        else:
            LOGGER.debug("Panel not paired")

            if not start_pairing:
                LOGGER.debug("Aborting pairing.")
                return False

            if not await self.start_initial_pairing():
                LOGGER.debug("Error Pairing with Panel")
                return False

        LOGGER.debug("Starting Plugin Operation")

        # Everything is configured
        return True

    async def start_operation(self) -> None:
        await self._task_manager.run(self.mqtt_connect_task(reconnect=True, run_forever=True), self._mqtt_task_connect_label)

    async def stop_operation(self) -> None:
        LOGGER.debug("Stopping Plugin Operation")

        if self.certificate_exchange_server is not None:
            self.certificate_exchange_server.close()

        if self.aiomqtt is not None:
            await self.aiomqtt.__aexit__(None, None, None)
            self.aiomqtt = None

        self._task_manager.cancel(self._mqtt_task_connect_label)
        self._task_manager.cancel(self._mqtt_task_listen_label)
        self._task_manager.cancel(self._mqtt_task_ping_label)
        self._task_manager.cancel(self._mqtt_task_config_label)

        self.connected = False
        self.connected_observer.notify()

    async def mqtt_connect_task(self, reconnect: bool, run_forever: bool) -> None:
        # Configure TLS parameters for MQTT connection
        tls_params = aiomqtt.TLSParameters(
            ca_certs=str(self._pki.qolsys_cer_file_path),
            certfile=str(self._pki.secure_file_path),
            keyfile=str(self._pki.key_file_path),
            cert_reqs=ssl.CERT_REQUIRED,
            tls_version=ssl.PROTOCOL_TLSv1_2,
            ciphers="ALL:@SECLEVEL=0",
        )

        LOGGER.debug("MQTT: Connecting ...")

        self._task_manager.cancel(self._mqtt_task_listen_label)
        self._task_manager.cancel(self._mqtt_task_ping_label)

        while True:
            try:
                self.aiomqtt = aiomqtt.Client(
                    hostname=self.settings.panel_ip,
                    port=8883,
                    tls_params=tls_params,
                    tls_insecure=True,
                    clean_session=True,
                    timeout=self.settings.mqtt_timeout,
                    identifier=self.settings.mqtt_remote_client_id,
                )

                await self.aiomqtt.__aenter__()

                LOGGER.info("MQTT: Client Connected")

                # Subscribe to panel internal database updates
                await self.aiomqtt.subscribe("iq2meid")

                # Subscribte to MQTT private response
                await self.aiomqtt.subscribe("response_" + self.settings.random_mac, qos=self.settings.mqtt_qos)

                # Subscribe to Z-Wave response
                await self.aiomqtt.subscribe("ZWAVE_RESPONSE", qos=self.settings.mqtt_qos)

                # Only log all traffic for debug purposes
                if self.settings.log_mqtt_mesages:
                    # Subscribe to MQTT commands send to panel by other devices
                    await self.aiomqtt.subscribe("mastermeid", qos=self.settings.mqtt_qos)

                self._task_manager.run(self.mqtt_listen_task(), self._mqtt_task_listen_label)
                self._task_manager.run(self.mqtt_ping_task(), self._mqtt_task_ping_label)

                response_connect = await self.command_connect()
                self.panel.imei = response_connect.get("master_imei", "")
                self.panel.product_type = response_connect.get("primary_product_type", "")

                await self.command_pingevent()
                await self.command_pair_status_request()

                response_database = await self.command_sync_database()
                LOGGER.debug("MQTT: Updating State from syncdatabase")
                self.panel.load_database(response_database.get("fulldbdata"))  # type: ignore[arg-type]
                self.panel.dump()
                self.state.dump()

                self.connected = True
                self.connected_observer.notify()

                if not run_forever:
                    self.connected = False
                    self.connected_observer.notify()
                    self._task_manager.cancel(self._mqtt_task_listen_label)
                    self._task_manager.cancel(self._mqtt_task_ping_label)
                    await self.aiomqtt.__aexit__(None, None, None)

                break

            except aiomqtt.MqttError as err:
                # Receive pannel network error
                self.connected = False
                self.connected_observer.notify()
                self.aiomqtt = None

                if reconnect:
                    LOGGER.debug("MQTT Error - %s: Connect - Reconnecting in %s seconds ...", err, self.settings.mqtt_timeout)
                    await asyncio.sleep(self.settings.mqtt_timeout)
                else:
                    raise QolsysMqttError from err

            except ssl.SSLError as err:
                # SSL error is and authentication error with invalid certificates en pki
                # We cannot recover from this error automaticly
                # Pannels need to be re-paired
                self.connected = False
                self.connected_observer.notify()
                self.aiomqtt = None
                raise QolsysSslError from err

    async def mqtt_ping_task(self) -> None:
        while True:
            if self.aiomqtt is not None and self.connected:
                with contextlib.suppress(aiomqtt.MqttError):
                    await self.command_pingevent()

            await asyncio.sleep(self.settings.mqtt_ping)

    async def mqtt_listen_task(self) -> None:
        try:
            async for message in self.aiomqtt.messages:  # type: ignore[union-attr]
                if self.settings.log_mqtt_mesages:  # noqa: SIM102
                    if isinstance(message.payload, bytes):
                        LOGGER.debug("MQTT TOPIC: %s\n%s", message.topic, message.payload.decode())

                # Panel response to MQTT Commands
                if message.topic.matches("response_" + self.settings.random_mac):  # noqa: SIM102
                    if isinstance(message.payload, bytes):
                        data = json.loads(message.payload.decode())
                        await self._mqtt_command_queue.handle_response(data)

                # Panel updates to IQ2MEID database
                if message.topic.matches("iq2meid"):  # noqa: SIM102
                    if isinstance(message.payload, bytes):
                        data = json.loads(message.payload.decode())
                        self.panel.parse_iq2meid_message(data)

                # Panel Z-Wave response
                if message.topic.matches("ZWAVE_RESPONSE"):  # noqa: SIM102
                    if isinstance(message.payload, bytes):
                        data = json.loads(message.payload.decode())
                        self.panel.parse_zwave_message(data)

        except aiomqtt.MqttError as err:
            self.connected = False
            self.connected_observer.notify()

            LOGGER.debug("%s: Listen - Reconnecting in %s seconds ...", err, self.settings.mqtt_timeout)
            await asyncio.sleep(self.settings.mqtt_timeout)
            self._task_manager.run(self.mqtt_connect_task(reconnect=True, run_forever=True), self._mqtt_task_connect_label)

    async def start_initial_pairing(self) -> bool:
        # check if random_mac exist
        if self.settings.random_mac == "":
            LOGGER.debug("Creating random_mac")
            self.settings.random_mac = generate_random_mac()
            self._pki.create(self.settings.random_mac, key_size=self.settings.key_size)

        # Check if PKI is valid
        self._pki.set_id(self.settings.random_mac)
        LOGGER.debug("Checking PKI")
        if not (self._pki.check_key_file() and self._pki.check_cer_file() and self._pki.check_csr_file()):
            LOGGER.error("PKI Error")
            return False

        LOGGER.debug("Starting Pairing Process")

        if not self.settings.check_plugin_ip():
            LOGGER.error("Plugin IP Address not configured")
            return False

        # If we dont allready have client signed certificate, start the pairing server
        if not self._pki.check_secure_file() or not self._pki.check_qolsys_cer_file() or not self.settings.check_panel_ip():
            # High Level Random Pairing Port
            pairing_port = random.randint(50000, 55000)

            # Start Pairing mDNS Brodcast
            LOGGER.debug("Starting mDNS Service Discovery: %s:%s", self.settings.plugin_ip, str(pairing_port))
            mdns_server = QolsysMDNS(self.settings.plugin_ip, pairing_port)
            await mdns_server.start_mdns()

            # Start Key Exchange Server
            LOGGER.debug("Starting Certificate Exchange Server")
            context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            context.load_cert_chain(certfile=self._pki.cer_file_path, keyfile=self._pki.key_file_path)
            self.certificate_exchange_server = await asyncio.start_server(
                self.handle_key_exchange_client, self.settings.plugin_ip, pairing_port, ssl=context
            )
            LOGGER.debug("Certificate Exchange Server Waiting for Panel")
            LOGGER.debug("Press Pair Button in IQ Remote Config Page ...")

            async with self.certificate_exchange_server:
                try:
                    await self.certificate_exchange_server.serve_forever()

                except asyncio.CancelledError:
                    LOGGER.debug("Stoping Certificate Exchange Server")
                    await self.certificate_exchange_server.wait_closed()
                    LOGGER.debug("Stoping mDNS Service Discovery")
                    await mdns_server.stop_mdns()

        LOGGER.debug("Sending MQTT Pairing Request to Panel")

        # We have client sgined certificate at this point
        # Connect to Panel MQTT to send pairing command
        await self._task_manager.run(self.mqtt_connect_task(reconnect=False, run_forever=False), self._mqtt_task_connect_label)
        LOGGER.debug("Plugin Pairing Completed ")
        return True

    async def handle_key_exchange_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:  # noqa: PLR0915
        received_panel_mac = False
        received_signed_client_certificate = False
        received_qolsys_cer = False

        try:
            continue_pairing = True
            while continue_pairing:
                # Plugin is receiving panel_mac from panel
                if not received_panel_mac and not received_signed_client_certificate and not received_qolsys_cer:
                    request = await reader.read(2048)
                    mac = request.decode()

                    address, port = writer.get_extra_info("peername")
                    LOGGER.debug("Panel Connected from: %s:%s", address, port)
                    LOGGER.debug("Receiving from Panel: %s", mac)

                    # Remove \x00 and \x01 from received string
                    self.settings.panel_mac = "".join(char for char in mac if char.isprintable())
                    self.settings.panel_ip = address
                    received_panel_mac = True

                    # Sending random_mac to panel
                    message = b"\x00\x11" + self.settings.random_mac.encode()
                    LOGGER.debug("Sending to Panel: %s", message.decode())
                    writer.write(message)
                    await writer.drain()

                    # Sending CSR File to panel
                    async with aiofiles.open(self._pki.csr_file_path, mode="rb") as f:
                        content = await f.read()
                        LOGGER.debug("Sending to Panel: [CSR File Content]")
                        writer.write(content)
                        writer.write(b"sent")
                        await writer.drain()

                    continue

                # Read signed certificate data
                if received_panel_mac and not received_signed_client_certificate and not received_qolsys_cer:
                    request = await reader.readuntil(b"sent")
                    if request.endswith(b"sent"):
                        request = request[:-4]

                    LOGGER.debug("Saving [Signed Client Certificate]")
                    async with aiofiles.open(self._pki.secure_file_path, mode="wb") as f:
                        await f.write(request)
                        received_signed_client_certificate = True

                # Read qolsys certificate data
                if received_panel_mac and received_signed_client_certificate and not received_qolsys_cer:
                    request = await reader.readuntil(b"sent")
                    if request.endswith(b"sent"):
                        request = request[:-4]

                    LOGGER.debug("Saving [Qolsys Certificate]")
                    async with aiofiles.open(self._pki.qolsys_cer_file_path, mode="wb") as f:
                        await f.write(request)
                        received_qolsys_cer = True
                        continue_pairing = False

                    continue

        except asyncio.CancelledError:
            LOGGER.exception("Key Exchange Server asyncio CancelledError")

        except Exception:
            LOGGER.exception("Key Exchange Server error")

        finally:
            writer.close()
            await writer.wait_closed()
            if self.certificate_exchange_server:
                self.certificate_exchange_server.close()

    async def command_connect(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending connect command")

        dhcpInfo = {
            "ipaddress": "",
            "gateway": "",
            "netmask": "",
            "dns1": "",
            "dns2": "",
            "dhcpServer": "",
            "leaseDuration": "",
        }

        command = MQTTCommand(self, "connect_v204")
        command.append("ipAddress", self.settings.plugin_ip)
        command.append("pairing_request", True)
        command.append("macAddress", self.settings.random_mac)
        command.append("remoteClientID", self.settings.mqtt_remote_client_id)
        command.append("softwareVersion", "4.4.1")
        command.append("productType", "tab07_rk68")
        command.append("bssid", "")
        command.append("lastUpdateChecksum", "2132501716")
        command.append("dealerIconsCheckSum", "")
        command.append("remote_feature_support_version", "1")
        command.append("current_battery_status", "Normal")
        command.append("remote_panel_battery_percentage", 100)
        command.append("remote_panel_battery_temperature", 430)
        command.append("remote_panel_battery_status", 3)
        command.append("remote_panel_battery_scale", 100)
        command.append("remote_panel_battery_voltage", 4102)
        command.append("remote_panel_battery_present", True)
        command.append("remote_panel_battery_technology", "")
        command.append("remote_panel_battery_level", 100)
        command.append("remote_panel_battery_health", 2)
        command.append("remote_panel_plugged", 1)
        command.append("dhcpInfo", json.dumps(dhcpInfo))

        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving connect command")
        return response

    async def command_pairing_request(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending pairing_request command")
        command = MQTTCommand(self, "connect_v204")

        dhcpInfo = {
            "ipaddress": "",
            "gateway": "",
            "netmask": "",
            "dns1": "",
            "dns2": "",
            "dhcpServer": "",
            "leaseDuration": "",
        }

        command.append("pairing_request", True)
        command.append("ipAddress", self.settings.plugin_ip)
        command.append("macAddress", self.settings.random_mac)
        command.append("remoteClientID", self.settings.mqtt_remote_client_id)
        command.append("softwareVersion", "4.4.1")
        command.append("productType", "tab07_rk68")
        command.append("bssid", "")
        command.append("lastUpdateChecksum", "2132501716")
        command.append("dealerIconsCheckSum", "")
        command.append("remote_feature_support_version", "1")
        command.append("dhcpInfo", json.dumps(dhcpInfo))

        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving pairing_request command")
        return response

    async def command_pingevent(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending pingevent command")
        command = MQTTCommand(self, "pingevent")
        command.append("remote_panel_status", "Active")
        command.append("macAddress", self.settings.random_mac)
        command.append("ipAddress", self.settings.plugin_ip)
        command.append("current_battery_status", "Normal")
        command.append("remote_panel_battery_percentage", 100)
        command.append("remote_panel_battery_temperature", 430)
        command.append("remote_panel_battery_status", 3)
        command.append("remote_panel_battery_scale", 100)
        command.append("remote_panel_battery_voltage", 4102)
        command.append("remote_panel_battery_present", True)
        command.append("remote_panel_battery_technology", "")
        command.append("remote_panel_battery_level", 100)
        command.append("remote_panel_battery_health", 2)
        command.append("remote_panel_plugged", 1)

        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving pingevent command")
        return response

    async def command_timesync(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending timeSync command")
        command = MQTTCommand(self, "timeSync")
        command.append("startTimestamp", int(time.time()))
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving timeSync command")
        return response

    async def command_sync_database(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending syncdatabase command")
        command = MQTTCommand(self, "syncdatabase")
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving syncdatabase command")
        return response

    async def command_acstatus(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending acStatus command")
        command = MQTTCommand(self, "acStatus")
        command.append("acStatus", "Connected")
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving acStatus command")
        return response

    async def command_dealer_logo(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending dealerLogo command")
        command = MQTTCommand(self, "dealerLogo")
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving dealerLogo command")
        return response

    async def command_pair_status_request(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending pair_status_request command")
        command = MQTTCommand(self, "pair_status_request")
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving pair_status_request command")
        return response

    async def command_disconnect(self) -> dict[str, Any]:
        LOGGER.debug("MQTT: Sending disconnect command")
        command = MQTTCommand(self, "disconnect")
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving disconnect command")
        return response

    async def command_ui_delay(self, partition_id: str, silent_disarming: bool = False) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending ui_delay command")
        command = MQTTCommand_Panel(self)

        # partition state needs to be sent for ui_delay to work
        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.error("command_ui_delay error: invalid partition %s", partition_id)
            return None

        arming_command = {
            "operation_name": "ui_delay",
            "panel_status": partition.system_status,
            "userID": 0,
            "partitionID": partition_id,  # STR EXPECTED
            "silentDisarming": silent_disarming,
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipcRequest = [
            {
                "dataType": "string",
                "dataValue": json.dumps(arming_command),
            }
        ]

        command.append("ipcRequest", ipcRequest)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving ui_delay command")
        return response

    async def command_disarm(
        self, partition_id: str, user_code: str = "", silent_disarming: bool = False
    ) -> dict[str, Any] | None:
        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.error("MQTT: disarm command error - Unknow Partition")
            return None

        # Do local user code verification
        user_id = 1
        if self.settings.check_user_code_on_disarm:
            user_id = self.panel.check_user(user_code)
            if user_id == -1:
                LOGGER.debug("MQTT: disarm command error - user_code error")
                raise QolsysUserCodeError()

        async def get_mqtt_disarm_command(silent_disarming: bool) -> str:
            if partition.alarm_state == PartitionAlarmState.ALARM:
                return "disarm_from_emergency"
            if partition.system_status in {
                PartitionSystemStatus.ARM_AWAY_EXIT_DELAY,
                PartitionSystemStatus.ARM_STAY_EXIT_DELAY,
                PartitionSystemStatus.ARM_NIGHT_EXIT_DELAY,
            }:
                return "disarm_from_openlearn_sensor"
            if partition.system_status in {
                PartitionSystemStatus.ARM_AWAY,
                PartitionSystemStatus.ARM_STAY,
                PartitionSystemStatus.ARM_NIGHT,
            }:
                await self.command_ui_delay(partition_id, silent_disarming)
                return "disarm_the_panel_from_entry_delay"

            return "disarm_from_openlearn_sensor"

        mqtt_disarm_command = await get_mqtt_disarm_command(silent_disarming)
        LOGGER.debug("MQTT: Sending disarm command - check_user_code:%s", self.settings.check_user_code_on_disarm)

        disarm_command = {
            "operation_name": mqtt_disarm_command,
            "userID": user_id,
            "partitionID": int(partition_id),  # INT EXPECTED
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(disarm_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving disarm command")
        return response

    async def command_arm(
        self,
        partition_id: str,
        arming_type: PartitionArmingType,
        user_code: str = "",
        exit_sounds: bool = False,
        instant_arm: bool = False,
        entry_delay: bool = True,
    ) -> dict[str, str] | None:
        LOGGER.debug(
            "MQTT: Sending arm command: partition%s, arming_type:%s, exit_sounds:%s, instant_arm: %s, entry_delay:%s",
            partition_id,
            arming_type,
            exit_sounds,
            instant_arm,
            entry_delay,
        )

        user_id = 0

        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.debug("MQTT: arm command error - Unknow Partition")
            return None

        if self.settings.check_user_code_on_arm:
            # Do local user code verification to arm
            user_id = self.panel.check_user(user_code)
            if user_id == -1:
                LOGGER.debug("MQTT: arm command error - user_code error")
                raise QolsysUserCodeError()

        exitSoundValue = "ON"
        if not exit_sounds:
            exitSoundValue = "OFF"

        entryDelay = "ON"
        if not entry_delay:
            entryDelay = "OFF"

        arming_command = {
            "operation_name": arming_type,
            "bypass_zoneid_set": "[]",
            "userID": user_id,
            "partitionID": int(partition_id),  # Expect Int
            "exitSoundValue": exitSoundValue,
            "entryDelayValue": entryDelay,
            "multiplePartitionsSelected": False,
            "instant_arming": instant_arm,
            "final_exit_arming_selected": False,
            "manually_selected_zones": "[]",
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(arming_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving arm command: partition%s", partition_id)
        return response

    async def command_panel_execute_scene(self, scene_id: str) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending execute_scene command")
        scene = self.state.scene(scene_id)
        if not scene:
            LOGGER.debug("MQTT: command_execute_scene Erro - Unknow Scene: %s", scene_id)
            return None

        scene_command = {
            "operation_name": "execute_scene",
            "scene_id": scene.scene_id,
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(scene_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving execute_scene command")
        return response

    async def command_panel_trigger_police(self, partition_id: str, silent: bool) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending panel_trigger_police command")

        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.debug("MQTT: command_panel_trigger_police Error - Unknow Partition: %s", partition_id)
            return None

        trigger_command = {
            "operation_name": "generate_emergency",
            "partitionID": int(partition_id),
            "zoneID": int(self._zone_id),
            "emergencyType": "Silent Police Emergency" if silent else "Police Emergency",
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(trigger_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving panel_trigger_police command")
        return response

    async def command_panel_trigger_auxilliary(self, partition_id: str, silent: bool) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending panel_trigger_auxilliary command")

        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.debug("MQTT: command_panel_trigger_auxilliary Error - Unknow Partition: %s", partition_id)
            return None

        trigger_command = {
            "operation_name": "generate_emergency",
            "partitionID": int(partition_id),
            "zoneID": int(self._zone_id),
            "emergencyType": "Silent Auxiliary Emergency" if silent else "Auxiliary Emergency",
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(trigger_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving panel_trigger_auxilliary command")
        return response

    async def command_panel_trigger_fire(self, partition_id: str) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending panel_trigger_fire command")

        partition = self.state.partition(partition_id)
        if not partition:
            LOGGER.debug("MQTT: command_panel_trigger_fire Error - Unknow Partition: %s", partition_id)
            return None

        trigger_command = {
            "operation_name": "generate_emergency",
            "partitionID": int(partition_id),
            "zoneID": int(self._zone_id),
            "emergencyType": "Fire Emergency",
            "operation_source": 1,
            "macAddress": self.settings.random_mac,
        }

        ipc_request = [
            {
                "dataType": "string",
                "dataValue": json.dumps(trigger_command),
            }
        ]

        command = MQTTCommand_Panel(self)
        command.append_ipc_request(ipc_request)
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving panel_trigger_fire command")
        return response

    async def command_zwave_switch_binary_set(self, node_id: str, status: bool) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending set_zwave_switch_binary command  - Node(%s) - Status(%s)", node_id, status)
        zwave_node = self.state.zwave_device(node_id)

        if not zwave_node:
            LOGGER.error("switch_binary_set - Invalid node_id %s", node_id)
            return None

        if zwave_node.generic_device_type not in (ZwaveDeviceClass.SwitchBinary, ZwaveDeviceClass.RemoteSwitchBinary):
            LOGGER.error("switch_binary_set used on invalid %s", zwave_node.generic_device_type)
            return None

        level = 0
        if status:
            level = 255

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.SwitchBinary, 1, level])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving set_zwave_switch_binary command")
        return response

    async def command_zwave_switch_multilevel_set(self, node_id: str, level: int) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending switch_multilevel_set command  - Node(%s) - Level(%s)", node_id, level)

        zwave_node = self.state.zwave_device(node_id)
        if not zwave_node:
            LOGGER.error("switch_multilevel_set - Invalid node_id %s", node_id)
            return None

        if zwave_node.generic_device_type not in (ZwaveDeviceClass.SwitchMultilevel, ZwaveDeviceClass.RemoteSwitchMultilevel):
            LOGGER.error("switch_multilevel_set used on invalid %s", zwave_node.generic_device_type)
            return None

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.SwitchMultilevel, 1, level])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving set_zwave_multilevel_switch command")
        return response

    async def command_zwave_doorlock_set(self, node_id: str, locked: bool) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending zwave_doorlock_set command - Node(%s) - Locked(%s)", node_id, locked)

        zwave_node = self.state.zwave_device(node_id)
        if not zwave_node:
            LOGGER.error("doorlock_set - Invalid node_id %s", node_id)
            return None

        # 0 unlocked, 255 locked
        lock_mode = 0
        if locked:
            lock_mode = 255

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.DoorLock, 1, lock_mode])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving zwave_doorlock_set command")
        return response

    async def command_zwave_thermostat_setpoint_set(
        self, node_id: str, mode: ThermostatMode, setpoint: float
    ) -> dict[str, Any] | None:
        LOGGER.debug(
            "MQTT: Sending zwave_thermostat_setpoint_set - Node(%s) - Mode(%s) - Setpoint(%s)", node_id, mode, setpoint
        )

        zwave_node = self.state.zwave_device(node_id)
        if not zwave_node:
            LOGGER.error("thermostat_setpoint_set - Invalid node_id %s", node_id)
            return None

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.ThermostatSetPoint, 1, mode, int(setpoint)])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving zwave_thermostat_mode_set command")
        return response

    async def command_zwave_thermostat_mode_set(self, node_id: str, mode: ThermostatMode) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending zwave_thermostat_mode_set command - Node(%s) - Mode(%s)", node_id, mode)

        zwave_node = self.state.zwave_device(node_id)
        if not zwave_node:
            LOGGER.error("thermostat_mode_set - Invalid node_id %s", node_id)
            return None

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.ThermostatMode, 1, mode])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving zwave_thermostat_mode_set command")
        return response

    async def command_zwave_thermostat_fan_mode_set(self, node_id: str, fan_mode: ThermostatFanMode) -> dict[str, Any] | None:
        LOGGER.debug("MQTT: Sending zwave_thermostat_fan_mode_set command - Node(%s) - FanMode(%s)", node_id, fan_mode)

        zwave_node = self.state.zwave_device(node_id)
        if not zwave_node:
            LOGGER.error("thermostat_fan_mode_set - Invalid node_id %s", node_id)
            return None

        command = MQTTCommand_ZWave(self, node_id, [ZwaveCommand.ThermostatFanMode, 1, fan_mode])
        response = await command.send_command()
        LOGGER.debug("MQTT: Receiving zwave_thermostat_fan_mode_set command")
        return response



import logging

from .observable import QolsysObservable

LOGGER = logging.getLogger(__name__)


class QolsysForecast:
    def __init__(self, data: dict[str, str]) -> None:
        self._high_temp: str = data.get("high_temp", "")
        self._low_temp: str = data.get("low_temp", "")
        self._day_of_week: str = data.get("day_of_week", "")
        self._condition: str = data.get("condition", "")
        self._icon: str = data.get("icon", "")
        self._precipitation: str = data.get("precipitation", "")
        self._current_weather_date: str = data.get("current_weather_date", "")

    @property
    def high_temp(self) -> float | None:
        try:
            return float(self._high_temp)
        except ValueError:
            return None

    @property
    def low_temp(self) -> float | None:
        try:
            return float(self._low_temp)
        except ValueError:
            return None

    @property
    def day_of_week(self) -> str:
        return self._day_of_week

    @property
    def condition(self) -> str:
        return self._condition

    @property
    def precipitation(self) -> int | None:
        try:
            return int(self._precipitation)
        except ValueError:
            return None

    @property
    def current_weather_date(self) -> str:
        return self._current_weather_date


class QolsysWeather(QolsysObservable):
    def __init__(self) -> None:
        super().__init__()
        self._forecasts: list[QolsysForecast] = []

    def current_weather(self) -> QolsysForecast | None:
        if self._forecasts:
            return self._forecasts[0]
        return None

    def update(self, data: list[QolsysForecast]) -> None:
        self._forecasts.clear()
        for forecast_data in data:
            self._forecasts.append(forecast_data)

        self.notify()

    @property
    def forecasts(self) -> list[QolsysForecast]:
        return self._forecasts



import json
import logging
import uuid
from typing import TYPE_CHECKING, Any

from .errors import QolsysMqttError

if TYPE_CHECKING:
    import aiomqtt

    from .controller import QolsysController

LOGGER = logging.getLogger(__name__)


class MQTTCommand:
    def __init__(
        self,
        controller: "QolsysController",
        eventName: str,
    ) -> None:
        self._controller: QolsysController = controller
        self._client: aiomqtt.Client | None = controller.aiomqtt
        self._topic: str = "mastermeid"
        self._eventName: str = eventName
        self._payload: dict[str, Any] = {}
        self._requestID = str(uuid.uuid4())
        self._qos: int = self._controller.settings.mqtt_qos
        self._responseTopic = "response_" + self._controller.settings.random_mac

        self.append("requestID", self._requestID)
        self.append("responseTopic", self._responseTopic)
        self.append("eventName", self._eventName)
        self.append("remoteMacAddress", self._controller.settings.random_mac)

    def append(self, argument: str, value: str | dict[str, Any] | int | bool | list[dict[str, Any]] | Any) -> None:
        self._payload[argument] = value

    async def send_command(self) -> dict[str, Any]:
        if self._client is None:
            LOGGER.error("MQTT Client not configured")
            raise QolsysMqttError

        await self._client.publish(topic=self._topic, payload=json.dumps(self._payload), qos=self._qos)
        return await self._controller.mqtt_command_queue.wait_for_response(self._requestID)


class MQTTCommand_IpcCall(MQTTCommand):
    def __init__(
        self,
        controller: "QolsysController",
        ipc_service_name: str,
        ipc_interface_name: str,
        ipc_transaction_id: int,
    ) -> None:
        super().__init__(controller, "ipcCall")
        self.append("ipcServiceName", ipc_service_name)
        self.append("ipcInterfaceName", ipc_interface_name)
        self.append("ipcTransactionID", ipc_transaction_id)

    def append_ipc_request(self, ipc_request: list[dict[str, Any]]) -> None:
        self.append("ipcRequest", ipc_request)


class MQTTCommand_Panel(MQTTCommand_IpcCall):
    def __init__(
        self,
        controller: "QolsysController",
    ) -> None:
        super().__init__(
            controller=controller,
            ipc_service_name="qinternalservice",
            ipc_interface_name="android.os.IQInternalService",
            ipc_transaction_id=7,
        )


class MQTTCommand_ZWave(MQTTCommand_IpcCall):
    def __init__(
        self,
        controller: "QolsysController",
        node_id: str,
        zwave_command: list[int],
    ) -> None:
        super().__init__(
            controller=controller,
            ipc_service_name="qzwaveservice",
            ipc_interface_name="android.os.IQZwaveService",
            ipc_transaction_id=47,
        )

        ipc_request: list[dict[str, Any]] = [
            {
                "dataType": "int",
                "dataValue": int(node_id),
            },
            {
                "dataType": "int",
                "dataValue": 0,
            },
            {
                "dataType": "byteArray",
                "dataValue": zwave_command,
            },
            {
                "dataType": "int",
                "dataValue": 0,
            },
            {
                "dataType": "int",
                "dataValue": 106,
            },
            {
                "dataType": "byteArray",
                "dataValue": [0],
            },
        ]

        self.append_ipc_request(ipc_request)



import asyncio
import logging

from .enum import DeviceCapability, ZoneSensorGroup, ZoneSensorType, ZoneStatus
from .observable import QolsysObservable
from .settings import QolsysSettings

LOGGER = logging.getLogger(__name__)


class QolsysZone(QolsysObservable):
    def __init__(self, data: dict[str, str], settings: QolsysSettings) -> None:
        super().__init__()

        self._settings = settings
        self._delay_task: asyncio.Task[None] | None = None

        self._zone_id: str = data.get("zoneid", "")
        self._sensorname: str = data.get("sensorname", "")
        self._sensorstatus: ZoneStatus = ZoneStatus(data.get("sensorstatus", ""))
        self._sensortype = ZoneSensorType(data.get("sensortype", ""))
        self._sensorgroup: str = data.get("sensorgroup", "")
        self._battery_status: str = data.get("battery_status", "")
        self._averagedBm: str = data.get("averagedBm", "")
        self._latestdBm: str = data.get("latestdBm", "")
        self._ac_status: str = data.get("ac_status", "")
        self._shortID: str = data.get("shortID", "")
        self._device_capability: str = data.get("device_capability", "")
        self._current_capability: str = data.get("current_capability", "")

        self._id: str = data.get("_id", "")
        self._zone_type: str = data.get("zone_type", "")
        self._sensor_id: str = data.get("sensorid", "")
        self._sensorstate: str = data.get("sensorstate", "")
        self._zone_physical_type: str = data.get("zone_physical_type", "")
        self._zone_alarm_type: str = data.get("zone_alarm_type", "")
        self._partition_id: str = data.get("partition_id", "")
        self._sensortts: str = data.get("sensortts", "")
        self._zone_rf_sensor: str = data.get("zone_rf_sensor", "")
        self._zone_supervised: str = data.get("zone_supervised", "")
        self._zone_reporting_enabled: str = data.get("zone_reporting_enabled", "")
        self._zone_two_way_voice_enabled: str = data.get("zone_two_way_voice_enabled", "")
        self._signal_source: str = data.get("signal_source", "")
        self._serial_number: str = data.get("serial_number", "")
        self._chimetype: str = data.get("chimetype", "")
        self._frame_count: str = data.get("frame_count", "")
        self._frame_type: str = data.get("frame_type", "")
        self._frame_id: str = data.get("frame_id", "")
        self._allowdisarming: str = data.get("allowdisarming", "")
        self._time: str = data.get("time", "")
        self._version: str = data.get("version", "")
        self._opr: str = data.get("opr", "")
        self._zone_equipement_code: str = data.get("zone_equipment_code", "")
        self._created_date: str = data.get("created_date", "")
        self._created_by: str = data.get("created_by", "")
        self._updated_by: str = data.get("updated_by", "")
        self._updated_date: str = data.get("updated_date", "")
        self._diag_24hr: str = data.get("diag_24hr", "")
        self._sub_type: str = data.get("sub_type", "")
        self._powerg_manufacture_id: str = data.get("powerg_manufacture_id", "")
        self._parent_node: str = data.get("parent_node", "")
        self._extras: str = data.get("extras", "")

        # EXTRA POWERG ATTRIBUTES
        self._powerg_long_id: str = ""
        self._powerg_status_data: str = ""
        self._powerg_temperature: str = ""
        self._powerg_light: str = ""
        self._powerg_notification_period: str = ""
        self._powerg_average_link_quality: str = ""
        self._powerg_link_quality: str = ""
        self._powerg_link_status: str = ""
        self._powerg_battery_voltage: str = ""

    def is_powerg_enabled(self) -> bool:
        return self._current_capability == DeviceCapability.POWERG

    def is_powerg_temperature_enabled(self) -> bool:
        return self._powerg_temperature != ""

    def is_powerg_light_enabled(self) -> bool:
        return self._powerg_light != ""

    def is_average_dbm_enabled(self) -> bool:
        return self.averagedBm is not None

    def is_latest_dbm_enabled(self) -> bool:
        return self.latestdBm is not None

    def is_battery_enabled(self) -> bool:
        return self.battery_status != ""

    def is_ac_enabled(self) -> bool:
        return self.ac_status != ""

    def update_powerg(self, data: dict[str, str]) -> None:
        short_id_update = data.get("shortID", "")
        if short_id_update != self.shortID:
            LOGGER.error(
                "Updating Zone%s PowerG Attribute (%s) with Zone%s (different shortID)",
                self._zone_id,
                self.sensorname,
                short_id_update,
            )
            return

        self.start_batch_update()

        if "longID" in data:
            self._powerg_long_id = data.get("longID", "")

        if "status_data" in data:
            self.powerg_status_data = data.get("status_data", "")

        if "temperature" in data:
            self.powerg_temperature = data.get("temperature", "")

        if "light" in data:
            self.powerg_light = data.get("light", "")

        if "notification_period" in data:
            self._powerg_notification_period = data.get("notification_period", "")

        if "average_link_quality" in data:
            self._powerg_average_link_quality = data.get("average_link_quality", "")

        if "link_quality" in data:
            self._powerg_link_quality = data.get("link_quality", "")

        if "link_status" in data:
            self._powerg_link_status = data.get("link_status", "")

        if "battery_voltage" in data:
            self._powerg_battery_voltage = data.get("battery_voltage", "")

        self.end_batch_update()

    def update(self, data: dict[str, str]) -> None:  # noqa: C901, PLR0912, PLR0915
        zone_id_update = data.get("zoneid", "")
        if zone_id_update != self._zone_id:
            LOGGER.error("Updating Zone%s (%s) with Zone%s (different id)", self._zone_id, self.sensorname, zone_id_update)
            return

        self.start_batch_update()

        if "sensorname" in data:
            self.sensorname = data.get("sensorname", "")

        if "sensorstatus" in data:
            self.sensorstatus = ZoneStatus(data.get("sensorstatus", ""))

        if "battery_status" in data:
            self.battery_status = data.get("battery_status", "")

        if "time" in data:
            self.time = data.get("time", "")

        if "partition_id" in data:
            self._partition_id = data.get("partition_id", "")

        if "lastestdBm" in data:
            self.latestdBm = data.get("latestdBm", "")

        if "averagedBm" in data:
            self.averagedBm = data.get("averagedBm", "")

        if "sensorgroup" in data:
            self.sensorgroup = data.get("sensorgroup", "")

        if "sensorstate" in data:
            self._sensorstate = data.get("sensorstate", "")

        if "sensortype" in data:
            self.sensortype = ZoneSensorType(data.get("sensortype", ""))

        if "zone_type" in data:
            self._zone_type = data.get("zone_type", "")

        if "zone_physical_type" in data:
            self._zone_physical_type = data.get("zone_physical_type", "")

        if "zone_alarm_type" in data:
            self._zone_alarm_type = data.get("zone_alarm_type", "")

        if "sensorttss" in data:
            self._sensortts = data.get("sensortts", "")

        if "current_capability" in data:
            self._current_capability = data.get("current_capability", "")

        if "zone_rf_sensor" in data:
            self._zone_rf_sensor = data.get("zone_rf_sensor", "")

        if "zone_supervised" in data:
            self._zone_supervised = data.get("zone_supervised", "")

        if "zone_reporting_enabled" in data:
            self._zone_reporting_enabled = data.get("zone_reporting_enabled", "")

        if "zone_two_way_voice_enabled" in data:
            self._zone_two_way_voice_enabled = data.get("zone_two_way_voice_enabled", "")

        if "signal_source" in data:
            self._signal_source = data.get("signal_source", "")

        if "serial_number" in data:
            self._serial_number = data.get("serial_number", "")

        if "chimetype" in data:
            self._chimetype = data.get("chimetype", "")

        if "frame_count" in data:
            self._frame_count = data.get("frame_count", "")

        if "frame_type" in data:
            self._frame_type = data.get("frame_type", "")

        if "allowdisarming" in data:
            self._allowdisarming = data.get("allowdisarming", "")

        self.end_batch_update()

    async def delay_zone(self, next_status: ZoneStatus) -> None:
        await asyncio.sleep(self._settings.motion_sensor_delay_sec)
        self._sensorstatus = next_status
        LOGGER.debug("Zone%s (%s) - sensorstatus: %s", self._zone_id, self.sensorname, next_status)
        self.notify()

    # -----------------------------
    # properties + setters
    # -----------------------------

    @property
    def id(self) -> str:
        return self._id

    @property
    def sensorname(self) -> str:
        return self._sensorname

    @sensorname.setter
    def sensorname(self, value: str) -> None:
        if self.sensorname != value:
            self._sensorname = value
            self.notify()

    @property
    def sensorgroup(self) -> str:
        return self._sensorgroup

    @sensorgroup.setter
    def sensorgroup(self, new_value: str) -> None:
        if self._sensorgroup != new_value:
            # Report new values
            try:
                ZoneSensorGroup(new_value)
            except ValueError:
                LOGGER.exception("Unknown Sensor group: %s, please report", new_value)

            self._sensorgroup = new_value
            LOGGER.debug("Zone%s (%s) - sensorgroup: %s", self.zone_id, self.sensorname, new_value)
            self.notify()

    @property
    def sensorstatus(self) -> ZoneStatus:
        return self._sensorstatus

    @sensorstatus.setter
    def sensorstatus(self, new_value: ZoneStatus) -> None:
        if self._settings.motion_sensor_delay and self._sensortype in {ZoneSensorType.MOTION, ZoneSensorType.PANEL_MOTION}:
            if new_value == ZoneStatus.IDLE:
                return
            if self._delay_task is not None:
                self._delay_task.cancel()
            self._delay_task = asyncio.create_task(self.delay_zone(ZoneStatus.IDLE))

        if self._sensorstatus != new_value:
            LOGGER.debug("Zone%s (%s) - sensorstatus: %s", self._zone_id, self.sensorname, new_value)
            self._sensorstatus = new_value
            self.notify()

    @property
    def battery_status(self) -> str:
        return self._battery_status

    @battery_status.setter
    def battery_status(self, value: str) -> None:
        if self._battery_status != value:
            LOGGER.debug("Zone%s (%s) - battery_status: %s", self.zone_id, self.sensorname, value)
            self._battery_status = value
            self.notify()

    @property
    def sensorstate(self) -> str:
        return self._sensorstate

    @property
    def sensortype(self) -> ZoneSensorType:
        return self._sensortype

    @sensortype.setter
    def sensortype(self, value: ZoneSensorType) -> None:
        if self._sensortype != value:
            self._sensortype = value
            self.notify()

    @property
    def zone_id(self) -> str:
        return self._zone_id

    @property
    def zone_type(self) -> str:
        return self._zone_type

    @property
    def zone_physical_type(self) -> str:
        return self._zone_physical_type

    @property
    def zone_alarm_type(self) -> str:
        return self._zone_alarm_type

    @property
    def partition_id(self) -> str:
        return self._partition_id

    @partition_id.setter
    def partition_id(self, value: str) -> None:
        if self.partition_id != value:
            self.partition_id = value
            self.notify()

    @property
    def shortID(self) -> str:
        return self._shortID

    @property
    def time(self) -> str:
        return self._time

    @time.setter
    def time(self, value: str) -> None:
        if self._time != value:
            self._time = value
            self.notify()

    @property
    def current_capability(self) -> str:
        return self._current_capability

    @property
    def latestdBm(self) -> int | None:
        try:
            n = int(self._latestdBm)
            if n >= 0 and n < 999:  # noqa: PLR2004
                return n
        except ValueError:
            pass
        return None

    @latestdBm.setter
    def latestdBm(self, value: str) -> None:
        if self._latestdBm != value:
            self.latestdBm = value
            self.notify()

    @property
    def averagedBm(self) -> int | None:
        try:
            n = int(self._averagedBm)
            if n >= 0 and n < 999:  # noqa: PLR2004
                return n
        except ValueError:
            pass
        return None

    @averagedBm.setter
    def averagedBm(self, value: str) -> None:
        if self._averagedBm != value:
            self._averagedBm = value
            self.notify()

    @property
    def device_capability(self) -> str:
        return self._device_capability

    @property
    def ac_status(self) -> str:
        return self._ac_status

    @property
    def powerg_temperature(self) -> float | None:
        try:
            temp = float(self._powerg_temperature)
            return round(temp, 1)
        except ValueError:
            return None

    @powerg_temperature.setter
    def powerg_temperature(self, value: str) -> None:
        if self._powerg_temperature != value:
            LOGGER.debug("Zone%s (%s) - powerg_temperature: %s", self._zone_id, self.sensorname, value)
            self._powerg_temperature = value
            self.notify()

    @property
    def powerg_light(self) -> float | None:
        try:
            return float(self._powerg_light)
        except ValueError:
            return None

    @powerg_light.setter
    def powerg_light(self, value: str) -> None:
        if self._powerg_light != value:
            LOGGER.debug("Zone%s (%s) - powerg_light: %s", self._zone_id, self.sensorname, value)
            self._powerg_light = value
            self.notify()

    @property
    def powerg_status_data(self) -> str:
        return self._powerg_status_data

    @powerg_status_data.setter
    def powerg_status_data(self, value: str) -> None:
        if self._powerg_status_data != value:
            LOGGER.debug("Zone%s (%s) - powerg_status_data: %s", self._zone_id, self.sensorname, value)
            self._powerg_status_data = value
            self.notify()

    def to_powerg_dict(self) -> dict[str, str]:
        return {
            "shortID": self.shortID,
            "longID": self._powerg_long_id,
            "status_data": self._powerg_status_data,
            "temperature": self._powerg_temperature,
            "light": self._powerg_light,
            "notification_period": self._powerg_notification_period,
            "average_link_quality": self._powerg_average_link_quality,
            "link_quality": self._powerg_link_quality,
            "link_status": self._powerg_link_status,
            "battery_voltage": self._powerg_battery_voltage,
        }

    def to_dict(self) -> dict[str, str]:
        return {
            "_id": self.id,
            "ac_status": self.ac_status,
            "allowdisarming": self._allowdisarming,
            "averagedBm": self._averagedBm,
            "battery_status": self.battery_status,
            "chimetype": self._chimetype,
            "created_by": self._created_by,
            "created_date": self._created_date,
            "current_capability": self._current_capability,
            "device_capability": self._device_capability,
            "diag_24hr": self._diag_24hr,
            "extras": self._extras,
            "sensorid": self._sensor_id,
            "sensorname": self.sensorname,
            "sensorgroup": self.sensorgroup,
            "sensorstatus": self.sensorstatus.value,
            "sensorstate": self.sensorstate,
            "sensortype": self.sensortype.value,
            "zoneid": self.zone_id,
            "zone_type": self.zone_type,
            "zone_physical_type": self.zone_physical_type,
            "zone_alarm_type": self.zone_alarm_type,
            "partition_id": self.partition_id,
            "sensortts": self._sensortts,
            "latestdBm": self._latestdBm,
            "zone_rf_sensor": self._zone_rf_sensor,
            "zone_supervised": self._zone_supervised,
            "zone_reporting_enabled": self._zone_reporting_enabled,
            "zone_two_way_voice_enabled": self._zone_two_way_voice_enabled,
            "signal_source": self._signal_source,
            "serial_number": self._serial_number,
            "frame_count": self._frame_count,
            "frame_type": self._frame_type,
            "frame_id": self._frame_id,
            "time": self.time,
            "version": self._version,
            "opr": self._opr,
            "zone_equipment_code": self._zone_equipement_code,
            "updated_by": self._updated_by,
            "updated_date": self._updated_date,
            "shortID": self._shortID,
            "sub_type": self._sub_type,
            "powerg_manufacture_id": self._powerg_manufacture_id,
            "parent_node": self._parent_node,
        }



import logging
from typing import Any

LOGGER = logging.getLogger(__name__)


class QolsysError(Exception):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class QolsysUserCodeError(QolsysError):
    def __init__(self) -> None:
        super().__init__("QolsysUserCodeError")


class QolsysSslError(QolsysError):
    def __init__(self) -> None:
        super().__init__("QolsysSslError")


class QolsysMqttError(QolsysError):
    def __init__(self) -> None:
        super().__init__("QolsysMqttError")


class QolsysSqlError(QolsysError):
    def __init__(self, operation: dict[str, Any]) -> None:
        super().__init__("QolsysSqlError")

        table = f"QolsysSqlError - table:{operation.get('table', '')}"
        query = f"QolsysSqlError - query:{operation.get('query', '')}"
        columns = f"QolsysSqlError - columns:{operation.get('columns', '')}"
        content_values = f"QolsysSqlError - content_values:{operation.get('content_value', '')}"
        selection = f"QolsysSqlError - selection:{operation.get('selection', '')}"
        selection_argument = f"QolsysSqlError - selection_argument:{operation.get('selection_argument', '')}"

        e = f"""\n{table}\n{query}\n{columns}\n{content_values}\n{selection}\n{selection_argument}"""
        LOGGER.exception(e)



import logging
import random

LOGGER = logging.getLogger(__name__)


def generate_random_mac() -> str:  # noqa: D103
    mac = [
        0xF2,
        0x16,
        0x3E,
        random.randint(0x00, 0x7F),
        random.randint(0x00, 0xFF),
        random.randint(0x00, 0xFF),
    ]
    return ":".join(map(lambda x: "%02x" % x, mac))  # noqa: C417, UP031



from enum import Enum, IntEnum


class vdFuncType(IntEnum):
    BINARY_ACTUATOR = 1
    MULTISTATE_ACTUATOR = 2
    LEVEL = 3
    STATUS = 4
    MOMENTARY_ACTION = 5
    MODE_SELECTOR = 6
    EVENT = 7


class vdFuncName(Enum):
    OPEN_CLOSE = "Open/Close"
    LOCK_UNLOCK = "Lock/Unlock"
    ON_OFF = "On/Off"


class vdFuncLocalControl(IntEnum):
    NONE = 0
    STATUS_ONLY = 1
    FULL_CONTROL = 2


class vdFuncState(IntEnum):
    ON = 0
    OFF = 1



from enum import Enum, IntEnum


class ThermostatMode(IntEnum):
    OFF = 0x0001
    HEAT = 0x0002
    COOL = 0x0004
    AUTO = 0x0008
    AUX_HEAT = 0x0010
    RESUME = 0x0020
    FAN_ONLY = 0x0040
    FURNACE = 0x0080
    DRY_AIR = 0x0100
    MOIST_AIR = 0x0200
    AUTO_CHANGEOVER = 0x0400
    ENERGY_SAVE_HEAT = 0x0800
    ENERGY_SAVE_COOL = 0x1000
    AWAY = 0x2000
    FULL_POWER = 0x4000
    MANUFACTURER_SPECEFIC = 0x8000


class ThermostatFanMode(IntEnum):
    AUTO_LOW = 0x0001
    LOW = 0x0002
    AUTO_HIGH = 0x0004
    HIGH = 0x0008
    AUTO_MEDIUM = 0x0010
    MEDIUM = 0x0020
    CIRCULATION = 0x4000
    HUMIDITY_CIRCULATION = 0x0080
    LEFT_RIGHT = 0x0100
    UP_DOWN = 0x0200
    QUIET = 0x0400
    EXTERNAL_CIRCULATION = 0x0800
    MANUFACTURER_SPECEFIC = 0x1000


class ZwaveCommand(IntEnum):
    SwitchBinary = 0x25
    SwitchMultilevel = 0x26
    SensorMultiLevel = 0x31
    ThermostatMode = 0x40
    ThermostatSetPoint = 0x43
    ThermostatFanMode = 0x44
    ThermostatFanState = 0x45
    DoorLock = 0x62
    Alarm = 0x71


class ZwaveDeviceClass(Enum):
    Unknown = 0x00
    GenericController = 0x01
    StaticController = 0x02
    AVControlPoint = 0x03
    Display = 0x04
    DoorLock = 0x05
    Thermostat = 0x06
    SensorBinary = 0x07
    SensorMultilevel = 0x08
    Meter = 0x09
    EntryControl = 0x0A
    SemiInteroperable = 0x0B
    Button = 0x0C
    RepeaterSlave = 0x0F
    SwitchBinary = 0x10
    SwitchMultilevel = 0x11
    RemoteSwitchBinary = 0x12
    RemoteSwitchMultilevel = 0x13
    SwitchToggleBinary = 0x14
    SwitchToggleMultilevel = 0x15
    ZIPNode = 0x16
    Ventilation = 0x17
    WindowCovering = 0x18
    BarrierOperator = 0x20
    SensorNotification = 0x21
    SoundSwitch = 0x22
    MeterPulse = 0x23
    ColorSwitch = 0x24
    ClimateControlSchedule = 0x25
    RemoteAssociationActivator = 0x26
    SceneController = 0x27
    SceneSceneActuatorConfiguration = 0x28
    SimpleAVControlPoint = 0x30



import logging
from typing import TYPE_CHECKING

from .adc_service import QolsysAdcService
from .enum_adc import vdFuncLocalControl, vdFuncName, vdFuncState, vdFuncType

if TYPE_CHECKING:
    from .adc_device import QolsysAdcDevice

LOGGER = logging.getLogger(__name__)


class QolsysAdcGarageDoorService(QolsysAdcService):
    def __init__(
        self,
        parent_device: "QolsysAdcDevice",
        id: int,
        func_name: vdFuncName,
        local_control: vdFuncLocalControl,
        func_type: vdFuncType,
        func_state: vdFuncState,
        timestamp: str,
    ) -> None:
        super().__init__(
            parent_device,
            id,
            func_name,
            local_control,
            func_type,
            func_state,
            timestamp,
        )

    def is_open(self) -> bool:
        return self.func_state == vdFuncState.ON



import asyncio
import logging
from collections.abc import Coroutine

LOGGER = logging.getLogger(__name__)


class QolsysTaskManager:
    def __init__(self) -> None:
        self._tasks: set[asyncio.Task] = set()  # type: ignore[type-arg]

    def run(self, coro: Coroutine, label: str) -> asyncio.Task:  # type: ignore[type-arg]
        task = asyncio.create_task(coro, name=label)
        self._tasks.add(task)

        def _done_callback(task: asyncio.Task) -> None:  # type: ignore[type-arg]
            try:
                task.result()

            except asyncio.CancelledError:
                LOGGER.debug("Task Cancelled: %s", task.get_name())

            except Exception:  # noqa: BLE001
                LOGGER.exception("[Callback] Task failed: %s", task.get_name())

            finally:
                self._tasks.discard(task)

        task.add_done_callback(_done_callback)
        return task

    def get_task(self, label: str) -> asyncio.Task | None:  # type: ignore[type-arg]
        for task in self._tasks:
            if task.get_name() == label:
                return task
        return None

    def cancel(self, label: str) -> None:
        for task in self._tasks:
            if task.get_name() == label:
                task.cancel()

    async def cancel_all(self) -> None:
        for task in self._tasks:
            task.cancel()
            await task

    async def wait_all(self) -> None:
        if self._tasks:
            await asyncio.gather(*self._tasks, return_exceptions=True)

    def dump(self) -> None:
        for task in self._tasks:
            LOGGER.debug("Task: %s, Done: %s, Cancelled: %s", task.get_name(), task.done(), task.cancelled())



import logging  # noqa: INP001
import sqlite3

from .table import QolsysTable

LOGGER = logging.getLogger(__name__)


class QolsysTablePowerGDevice(QolsysTable):
    def __init__(self, db: sqlite3.Connection, cursor: sqlite3.Cursor) -> None:
        super().__init__(db, cursor)
        self._uri = "content://com.qolsys.qolsysprovider.PowerGDeviceContentProvider/powerg_device"
        self._table = "powerg_device"
        self._abort_on_error = False
        self._implemented = True

        self._columns = [
            "_id",
            "avg_link_quality",
            "battery_voltage",
            "capabilities",
            "dealer_code",
            "extras",
            "firmware_version",
            "led",
            "light",
            "link_quality",
            "link_status",
            "longID",
            "manufacturing_id",
            "notification_period",
            "opr",
            "parent_node",
            "partition_id",
            "radio_id",
            "radio_version",
            "shortID",
            "status_data",
            "supported_type",
            "temperature",
            "version",
            "writeable_capabilities",
        ]

        self._create_table()



import logging
from typing import TYPE_CHECKING

from .enum_adc import vdFuncLocalControl, vdFuncName, vdFuncState, vdFuncType

if TYPE_CHECKING:
    from .adc_device import QolsysAdcDevice

LOGGER = logging.getLogger(__name__)


class QolsysAdcService:
    def __init__(
        self,
        parent_device: "QolsysAdcDevice",
        id: int,
        func_name: vdFuncName,
        local_control: vdFuncLocalControl,
        func_type: vdFuncType,
        func_state: vdFuncState,
        timestamp: str,
    ) -> None:
        super().__init__()
        self._parent_device: QolsysAdcDevice = parent_device
        self._id: int = id
        self._func_name: vdFuncName = func_name
        self._local_control: vdFuncLocalControl = local_control
        self._func_type: vdFuncType = func_type
        self._func_state: vdFuncState = func_state
        self._timestamp: str = timestamp

    def update_adc_service(
        self,
        func_name: vdFuncName,
        local_control: vdFuncLocalControl,
        func_type: vdFuncType,
        func_state: vdFuncState,
        timestamp: str,
    ) -> None:
        self.name = func_name
        self.local_control = local_control
        self.func_type = func_type
        self.func_state = func_state
        self.timestamp = timestamp

    # -----------------------------
    # properties + setters
    # -----------------------------

    @property
    def id(self) -> int:
        return self._id

    @property
    def func_name(self) -> vdFuncName:
        return self._func_name

    @func_name.setter
    def func_name(self, value: vdFuncName) -> None:
        if value != self._func_name:
            self._func_name = value
            LOGGER.debug(
                "ADC%s (%s) - Func%s - func_name:%s",
                self._parent_device.device_id,
                self._parent_device.name,
                self.id,
                self.func_name,
            )
            LOGGER.debug("Warning - Changing func_name will change device type - Not supported")
            self._parent_device.notify()

    @property
    def local_control(self) -> vdFuncLocalControl:
        return self._local_control

    @local_control.setter
    def local_control(self, value: vdFuncLocalControl) -> None:
        if value != self.local_control:
            self._local_control = value
            LOGGER.debug(
                "ADC%s (%s) - Func%s - local_control:%s",
                self._parent_device.device_id,
                self._parent_device.name,
                self.id,
                self.local_control,
            )
            self._parent_device.notify()

    @property
    def func_type(self) -> vdFuncType:
        return self._func_type

    @func_type.setter
    def func_type(self, value: vdFuncType) -> None:
        if value != self._func_type:
            self._func_type = value
            LOGGER.debug(
                "ADC%s (%s) - Func%s - func_type:%s",
                self._parent_device.device_id,
                self._parent_device.name,
                self.id,
                self.func_type,
            )
            LOGGER.debug("Warning - Changing func_type will change device type - Not supported")
            self._parent_device.notify()

    @property
    def func_state(self) -> vdFuncState:
        return self._func_state

    @func_state.setter
    def func_state(self, value: vdFuncState) -> None:
        if value != self._func_state:
            self._func_state = value
            LOGGER.debug(
                "ADC%s (%s) - Func%s - func_state:%s",
                self._parent_device.device_id,
                self._parent_device.name,
                self.id,
                self.func_state,
            )
            self._parent_device.notify()

    @property
    def timestamp(self) -> str:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: str) -> None:
        if value != self._timestamp:
            self._timestamp = value
            LOGGER.debug(
                "ADC%s (%s) - Func%s - timestamp:%s",
                self._parent_device.device_id,
                self._parent_device.name,
                self.id,
                self.timestamp,
            )
            self._parent_device.notify()



from __future__ import annotations

import base64
import json
import logging
from typing import TYPE_CHECKING, Any

from qolsys_controller.adc_device import QolsysAdcDevice
from qolsys_controller.zwave_power import QolsysPower
from qolsys_controller.zwave_thermometer import QolsysThermometer

from .database.db import QolsysDB
from .enum import (
    PartitionAlarmState,
    PartitionAlarmType,
    PartitionSystemStatus,
)
from .observable import QolsysObservable
from .partition import QolsysPartition
from .scene import QolsysScene
from .users import QolsysUser
from .weather import QolsysForecast, QolsysWeather
from .zone import QolsysZone
from .zwave_dimmer import QolsysDimmer
from .zwave_generic import QolsysGeneric
from .zwave_lock import QolsysLock
from .zwave_thermostat import QolsysThermostat

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .controller import QolsysController
    from .zwave_device import QolsysZWaveDevice


class QolsysPanel(QolsysObservable):
    def __init__(self, controller: QolsysController) -> None:
        self._controller = controller
        self._db = QolsysDB()

        # Partition settings
        self.settings_partition = ["SYSTEM_STATUS", "EXIT_SOUNDS", "ENTRY_DELAYS", "SYSTEM_STATUS_CHANGED_TIME"]
        self.state_partition = ["ALARM_STATE"]

        # Panel settings
        self.settings_panel_observer = QolsysObservable()
        self.settings_panel = [
            "PANEL_TAMPER_STATE",
            "AC_STATUS",
            "BATTERY_STATUS",
            "FAIL_TO_COMMUNICATE",
            "SECURE_ARMING",
            "AUTO_BYPASS",
            "AUTO_STAY",
            "AUTO_ARM_STAY",
            "AUTO_EXIT_EXTENSION",
            "FINAL_EXIT_DOOR_ARMING",
            "NO_ARM_LOW_BATTERY",
            "TEMPFORMAT",
            "LANGUAGE",
            "COUNTRY",
            "SYSTEM_TIME",
            "GSM_CONNECTION_STATUS",
            "GSM_SIGNAL_STRENGTH",
            "ANDROID_VERSION",
            "HARDWARE_VERSION",
            "TIMER_NORMAL_ENTRY_DELAY",
            "TIMER_NORMAL_EXIT_DELAY",
            "TIMER_LONG_ENTRY_DELAY",
            "TIMER_LONG_EXIT_DELAY",
            "ZWAVE_CONTROLLER",
            "ZWAVE_CARD",
            "POLICE_PANIC_ENABLED",
            "FIRE_PANIC_ENABLED",
            "AUXILIARY_PANIC_ENABLED",
            "NIGHTMODE_SETTINGS",
            "NIGHT_SETTINGS_STATE",
            "PARTITIONS",
            "SIX_DIGIT_USER_CODE",
            "SHOW_SECURITY_SENSORS",
            "SYSTEM_LOGGED_IN_USER",
            "PANEL_SCENES_SETTING",
            "CONTROL_4",
            "ZWAVE_FIRM_WARE_VERSION",
            "FINAL_EXIT_DOOR_ARMING",
            "NO_ARM_LOW_BATTERY",
            "MAC_ADDRESS",
            "LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N",
        ]

        self._PANEL_TAMPER_STATE: str = ""
        self._AC_STATUS: str = ""
        self._BATTERY_STATUS: str = ""
        self._FAIL_TO_COMMUNICATE: str = ""
        self._SECURE_ARMING: str = ""
        self._AUTO_BYPASS: str = ""
        self._AUTO_STAY: str = ""
        self._AUTO_ARM_STAY: str = ""
        self._AUTO_EXIT_EXTENSION: str = ""
        self._FINAL_EXIT_DOOR_ARMING: str = ""
        self._NO_ARM_LOW_BATTERY: str = ""
        self._TEMPFORMAT: str = ""
        self._LANGUAGE: str = ""
        self._COUNTRY: str = ""
        self._SYSTEM_TIME: str = ""
        self._GSM_CONNECTION_STATUS: str = ""
        self._GSM_SIGNAL_STRENGTH: str = ""
        self._ANDROID_VERSION: str = ""
        self._HARDWARE_VERSION: str = ""
        self._TIMER_NORMAL_ENTRY_DELAY: str = ""
        self._TIMER_NORMAL_EXIT_DELAY: str = ""
        self._TIMER_LONG_ENTRY_DELAY: str = ""
        self._TIMER_LONG_EXIT_DELAY: str = ""
        self._ZWAVE_FIRM_WARE_VERSION: str = ""
        self._ZWAVE_CONTROLLER: str = ""
        self._ZWAVE_CARD: str = ""
        self._POLICE_PANIC_ENABLED: str = ""
        self._FIRE_PANIC_ENABLED: str = ""
        self._AUXILIARY_PANIC_ENABfLED: str = ""
        self._NIGHTMODE_SETTINGS: str = ""
        self._NIGHT_SETTINGS_STATE: str = ""
        self._PARTITIONS: str = ""
        self._SIX_DIGIT_USER_CODE: str = ""
        self._SHOW_SECURITY_SENSORS: str = ""
        self._CONTROL_4: str = ""
        self._MAC_ADDRESS: str = ""
        self._SYSTEM_LOGGED_IN_USER: str = ""
        self._PANEL_SCENES_SETTING: str = ""
        self._LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N: str = ""

        self._users: list[QolsysUser] = []
        self._unique_id: str = ""

        self._imei: str = ""
        self._product_type: str = ""

    def read_users_file(self) -> bool:
        # Loading user_code data from users.conf file if exists
        if self._controller.settings.users_file_path.is_file():
            try:
                path = self._controller.settings.users_file_path
                with path.open("r", encoding="utf-8") as file:
                    try:
                        users = json.load(file)
                        for user in users:
                            qolsys_user = QolsysUser()
                            qolsys_user.id = user.get("id")
                            qolsys_user.user_code = user.get("user_code")
                            self._users.append(qolsys_user)

                    except json.JSONDecodeError:
                        LOGGER.exception("users.conf file json error")
                        return False

            except FileNotFoundError:
                LOGGER.exception("users.conf file not found")
                return False

        return True

    @property
    def db(self) -> QolsysDB:
        return self._db

    @property
    def ANDROID_VERSION(self) -> str:
        self._ANDROID_VERSION = self.db.get_setting_panel("ANDROID_VERSION")
        return self._ANDROID_VERSION

    @property
    def PANEL_TAMPER_STATE(self) -> str:
        self._PANEL_TAMPER_STATE = self.db.get_setting_panel("PANEL_TAMPER_STATE")
        return self._PANEL_TAMPER_STATE

    @property
    def AC_STATUS(self) -> str:
        self._AC_STATUS = self.db.get_setting_panel("AC_STATUS")
        return self._AC_STATUS

    @property
    def BATTERY_STATUS(self) -> str:
        self._BATTERY_STATUS = self.db.get_setting_panel("BATTERY_STATUS")
        return self._BATTERY_STATUS

    @property
    def FAIL_TO_COMMUNICATE(self) -> str:
        self._FAIL_TO_COMMUNICATE = self.db.get_setting_panel("FAIL_TO_COMMUNICATE")
        return self._FAIL_TO_COMMUNICATE

    @property
    def SECURE_ARMING(self) -> str:
        self._SECURE_ARMING = self.db.get_setting_panel("SECURE_ARMING")
        return self._SECURE_ARMING

    @property
    def AUTO_BYPASS(self) -> str:
        self._AUTO_BYPASS = self.db.get_setting_panel("AUTO_BYPASS")
        return self._AUTO_BYPASS

    @property
    def AUTO_STAY(self) -> str:
        self._AUTO_STAY = self.db.get_setting_panel("AUTO_STAY")
        return self._AUTO_STAY

    @property
    def AUTO_ARM_STAY(self) -> str:
        self._AUTO_ARM_STAY = self.db.get_setting_panel("AUTO_ARM_STAY")
        return self._AUTO_ARM_STAY

    @property
    def AUTO_EXIT_EXTENSION(self) -> str:
        self._AUTO_EXIT_EXTENSION = self.db.get_setting_panel("AUTO_EXIT_EXTENSION")
        return self._AUTO_EXIT_EXTENSION

    @property
    def FINAL_EXIT_DOOR_ARMING(self) -> str:
        self._FINAL_EXIT_DOOR_ARMING = self.db.get_setting_panel("FINAL_EXIT_DOOR_ARMING")
        return self._FINAL_EXIT_DOOR_ARMING

    @property
    def NO_ARM_LOW_BATTERY(self) -> str:
        self._NO_ARM_LOW_BATTERY = self.db.get_setting_panel("NO_ARM_LOW_BATTERY")
        return self._NO_ARM_LOW_BATTERY

    @property
    def TEMPFORMAT(self) -> str:
        self._TEMPFORMAT = self.db.get_setting_panel("TEMPFORMAT")
        return self._TEMPFORMAT

    @property
    def LANGUAGE(self) -> str:
        self._LANGUAGE = self.db.get_setting_panel("LANGUAGE")
        return self._LANGUAGE

    @property
    def COUNTRY(self) -> str:
        self._COUNTRY = self.db.get_setting_panel("COUNTRY")
        return self._COUNTRY

    @property
    def SYSTEM_TIME(self) -> str:
        self._SYSTEM_TIME = self.db.get_setting_panel("SYSTEM_TIME")
        return self._SYSTEM_TIME

    @property
    def GSM_CONNECTION_STATUS(self) -> str:
        self._GSM_CONNECTION_STATUS = self.db.get_setting_panel("GSM_CONNECTION_STATUS")
        return self._GSM_CONNECTION_STATUS

    @property
    def GSM_SIGNAL_STRENGTH(self) -> str:
        self._GSM_SIGNAL_STRENGTH = self.db.get_setting_panel("GSM_SIGNAL_STRENGTH")
        return self._GSM_SIGNAL_STRENGTH

    @property
    def HARDWARE_VERSION(self) -> str:
        self._HARDWARE_VERSION = self.db.get_setting_panel("HARDWARE_VERSION")
        return self._HARDWARE_VERSION

    @property
    def ZWAVE_FIRM_WARE_VERSION(self) -> str:
        self._ZWAVE_FIRM_WARE_VERSION = self.db.get_setting_panel("ZWAVE_FIRM_WARE_VERSION")
        return self._ZWAVE_FIRM_WARE_VERSION

    @property
    def ZWAVE_CONTROLLER(self) -> str:
        self._ZWAVE_CONTROLLER = self.db.get_setting_panel("ZWAVE_CONTROLLER")
        return self._ZWAVE_CONTROLLER

    @property
    def ZWAVE_CARD(self) -> str:
        self._ZWAVE_CARD = self.db.get_setting_panel("ZWAVE_CARD")
        return self._ZWAVE_CARD

    @property
    def POLICE_PANIC_ENABLED(self) -> str:
        self._POLICE_PANIC_ENABLED = self.db.get_setting_panel("POLICE_PANIC_ENABLED")
        return self._POLICE_PANIC_ENABLED

    @property
    def FIRE_PANIC_ENABLED(self) -> str:
        self._FIRE_PANIC_ENABLED = self.db.get_setting_panel("FIRE_PANIC_ENABLED")
        return self._FIRE_PANIC_ENABLED

    @property
    def AUXILIARY_PANIC_ENABLED(self) -> str:
        self._AUXILIARY_PANIC_ENABLED = self.db.get_setting_panel("AUXILIARY_PANIC_ENABLED")
        return self._AUXILIARY_PANIC_ENABLED

    @property
    def PARTITIONS(self) -> str:
        self._PARTITIONS = self.db.get_setting_panel("PARTITIONS")
        return self._PARTITIONS

    @property
    def SIX_DIGIT_USER_CODE(self) -> str:
        self._SIX_DIGIT_USER_CODE = self.db.get_setting_panel("SIX_DIGIT_USER_CODE")
        return self._SIX_DIGIT_USER_CODE

    @property
    def CONTROL_4(self) -> str:
        self._CONTROL_4 = self.db.get_setting_panel("CONTROL_4")
        return self._CONTROL_4

    @property
    def NIGHTMODE_SETTINGS(self) -> str:
        self._NIGHTMODE_SETTINGS = self.db.get_setting_panel("NIGHTMODE_SETTINGS")
        return self._NIGHTMODE_SETTINGS

    @property
    def NIGHT_SETTINGS_STATE(self) -> str:
        self._NIGHT_SETTINGS_STATE = self.db.get_setting_panel("NIGHT_SETTINGS_STATE")
        return self._NIGHT_SETTINGS_STATE

    @property
    def SHOW_SECURITY_SENSORS(self) -> str:
        self._SHOW_SECURITY_SENSORS = self.db.get_setting_panel("SHOW_SECURITY_SENSORS")
        return self._SHOW_SECURITY_SENSORS

    @property
    def TIMER_NORMAL_ENTRY_DELAY(self) -> str:
        self._TIMER_NORMAL_ENTRY_DELAY = self.db.get_setting_panel("TIMER_NORMAL_ENTRY_DELAY")
        return self._TIMER_NORMAL_ENTRY_DELAY

    @property
    def TIMER_NORMAL_EXIT_DELAY(self) -> str:
        self._TIMER_NORMAL_EXIT_DELAY = self.db.get_setting_panel("TIMER_NORMAL_EXIT_DELAY")
        return self._TIMER_NORMAL_EXIT_DELAY

    @property
    def TIMER_LONG_ENTRY_DELAY(self) -> str:
        self._TIMER_LONG_ENTRY_DELAY = self.db.get_setting_panel("TIMER_LONG_ENTRY_DELAY")
        return self._TIMER_LONG_ENTRY_DELAY

    @property
    def MAC_ADDRESS(self) -> str:
        self._MAC_ADDRESS = self.db.get_setting_panel("MAC_ADDRESS")
        return self._MAC_ADDRESS

    @MAC_ADDRESS.setter
    def MAC_ADDRESS(self, value: str) -> None:
        self._MAC_ADDRESS = value

    @property
    def unique_id(self) -> str:
        mac_address = self.MAC_ADDRESS
        return mac_address.replace(":", "")

    @property
    def TIMER_LONG_EXIT_DELAY(self) -> str:
        self._TIMER_LONG_EXIT_DELAY = self.db.get_setting_panel("TIMER_LONG_ENTRY_DELAY")
        return self._TIMER_LONG_EXIT_DELAY

    @property
    def LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N(self) -> str:
        self._LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N = self.db.get_setting_panel("LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N")
        return self._LAST_UPDATE_IQ_REMOTE_PATCH_CKECKSUM_N

    @property
    def imei(self) -> str:
        return self._imei

    @imei.setter
    def imei(self, value: str) -> None:
        self._imei = value

    @property
    def product_type(self) -> str:
        return self._product_type

    @product_type.setter
    def product_type(self, value: str) -> None:
        self._product_type = value

    @property
    def SYSTEM_LOGGED_IN_USER(self) -> str:
        self._SYSTEM_LOGGED_IN_USER = self.db.get_setting_panel("SYSTEM_LOGGED_IN_USER")
        return self._SYSTEM_LOGGED_IN_USER

    @SYSTEM_LOGGED_IN_USER.setter
    def SYSTEM_LOGGED_IN_USER(self, value: str) -> None:
        self._SYSTEM_LOGGED_IN_USER = value

    @property
    def PANEL_SCENES_SETTING(self) -> str:
        self._PANEL_SCENES_SETTING = self.db.get_setting_panel("PANEL_SCENES_SETTING")
        return self.PANEL_SCENES_SETTING

    def load_database(self, database: list[dict[str, Any]]) -> None:
        self.db.load_db(database)
        self._controller.state.sync_partitions_data(self.get_partitions_from_db())
        self._controller.state.sync_zones_data(self.get_zones_from_db())
        self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())
        self._controller.state.sync_adc_devices_data(self.get_adc_devices_from_db())
        self._controller.state.sync_scenes_data(self.get_scenes_from_db())
        self._controller.state.sync_weather_data(self.get_weather_from_db())

        # Validate all local user match a Qolsys Panel user
        qolsys_users = self.db.get_users()
        qolsys_user_list = []
        for qolsys_user in qolsys_users:
            try:
                userid = int(qolsys_user.get("userid", ""))
            except ValueError:
                LOGGER.error("Invalid userid in panel database: %s", qolsys_user.get("userid", ""))
                userid = -1

            qolsys_user_list.append(userid)

        for local_user in self._users:
            if local_user.id not in qolsys_user_list:
                LOGGER.error("ID %s from users.conf file not found in panel database", local_user.id)

        # Check associated zone_id in iqremotesettins table
        LOGGER.debug("Checking iqremotesettings table for zone_id matching panel MAC address")
        iqremote_settings_list = self.db.get_iqremote_settings()
        for iqremote in iqremote_settings_list:
            if (
                self._controller.settings.random_mac.replace(":", "").lower()
                == iqremote.get("mac_address", "").replace(":", "").lower()
            ):
                self._controller._zone_id = iqremote.get("zone_id", "")
                LOGGER.debug("Found matching zone_id: %s", self._controller._zone_id)
                break

    # Parse Z-Wave message
    def parse_zwave_message(self, data: dict[str, Any]) -> None:
        zwave = data.get("ZWAVE_RESPONSE", "")
        decoded_payload = base64.b64decode(zwave.get("ZWAVE_PAYLOAD", "")).hex()

        LOGGER.debug(
            "Z-Wave Response: Node(%s) - Status(%s) - Payload(%s)",
            zwave.get("NODE_ID", ""),
            zwave.get("ZWAVE_COMMAND_STATUS", ""),
            decoded_payload,
        )

        node_id: str = str(zwave.get("NODE_ID", ""))
        command_status: int = zwave.get("ZWAVE_COMMAND_STATUS", "")

        if len(decoded_payload) % 2 != 0:
            LOGGER.error("Payload must have even length:%s", decoded_payload)
            return

        payload_bytes = bytes.fromhex(decoded_payload)
        command_class = int(payload_bytes[0])
        command_type = int(payload_bytes[1])
        value = list(payload_bytes[2:])

        node = self._controller.state.zwave_device(node_id)
        if node is not None:
            node.update_raw(command_class, command_status, command_type, value)

    # Parse panel update to database
    def parse_iq2meid_message(self, data: dict[str, Any]) -> None:  # noqa: C901, PLR0912, PLR0915
        eventName = data.get("eventName")
        dbOperation = data.get("dbOperation", "")
        uri = data.get("uri")

        match eventName:
            case "stopScreenCapture":
                pass

            case "primaryDisconnect":
                LOGGER.info("Main Panel Disconnect")

            case "dbChanged":
                match dbOperation:
                    case "update":
                        content_values = data.get("contentValues", "")
                        selection = data.get("selection")
                        selection_argument = data.get("selectionArgs")

                        match uri:
                            # Update Settings Content Provider
                            case self.db.table_qolsyssettings.uri:
                                name = content_values.get("name", "")
                                new_value = content_values.get("value", "")
                                old_value = self.db.get_setting_panel(name)
                                self.db.table_qolsyssettings.update(selection, selection_argument, content_values)

                                # Update Panel Settings - Send notification if settings ha changed
                                if name in self.settings_panel and old_value != new_value:
                                    LOGGER.debug("Panel Setting - %s: %s", name, new_value)
                                    self.settings_panel_observer.notify()

                                # Update Partition setting - Send notification if setting has changed
                                if name in self.settings_partition:
                                    partition_id = content_values.get("partition_id", "")
                                    partition = self._controller.state.partition(partition_id)
                                    if partition is not None:
                                        match name:
                                            case "SYSTEM_STATUS":
                                                partition.system_status = PartitionSystemStatus(new_value)
                                            case "SYSTEM_STATUS_CHANGED_TIME":
                                                partition.system_status_changed_time = new_value
                                            case "EXIT_SOUNDS":
                                                partition.exit_sounds = new_value
                                            case "ENTRY_DELAYS":
                                                partition.entry_delays = new_value

                            # Update Sensor Content Provider
                            case self.db.table_sensor.uri:
                                self.db.table_sensor.update(selection, selection_argument, content_values)
                                zoneid = content_values.get("zoneid", "")
                                zone = self._controller.state.zone(zone_id=zoneid)
                                if zone is not None:
                                    zone.update(content_values)

                            # Update State
                            case self.db.table_state.uri:
                                name = content_values.get("name", "")
                                new_value = content_values.get("value", "")
                                partition_id = content_values.get("partition_id", "")
                                self.db.table_state.update(selection, selection_argument, content_values)

                                if name in self.state_partition:
                                    partition = self._controller.state.partition(partition_id)
                                    if partition is not None:
                                        match name:
                                            case "ALARM_STATE":
                                                partition.alarm_state = PartitionAlarmState(new_value)

                            # Update heat_map
                            case self.db.table_heat_map.uri:
                                self.db.table_heat_map.update(selection, selection_argument, content_values)

                            # Update master_slave
                            case self.db.table_master_slave.uri:
                                self.db.table_master_slave.update(selection, selection_argument, content_values)

                            # Update dashboard_msgs
                            case self.db.table_dashboard_msgs.uri:
                                self.db.table_dashboard_msgs.update(selection, selection_argument, content_values)

                            # Update PartitionContentProvider
                            case self.db.table_partition.uri:
                                self.db.table_partition.update(selection, selection_argument, content_values)
                                partition_id = content_values.get("partition_id", "")
                                partition = self._controller.state.partition(partition_id)
                                if partition is not None:
                                    partition.update_partition(content_values)

                            # Update History Content Provider
                            case self.db.table_history.uri:
                                self.db.table_history.update(selection, selection_argument, content_values)

                            # Update DimmerLightsContentProvider
                            case self.db.table_dimmer.uri:
                                self.db.table_dimmer.update(selection, selection_argument, content_values)
                                node_id = content_values.get("node_id", "")
                                node = self._controller.state.zwave_device(node_id)
                                if node is not None and isinstance(node, QolsysDimmer):
                                    node.update_dimmer(content_values)

                            # Update Thermostat Content Provider
                            case self.db.table_thermostat.uri:
                                self.db.table_thermostat.update(selection, selection_argument, content_values)
                                node_id = content_values.get("node_id", "")
                                node = self._controller.state.zwave_device(node_id)
                                if node is not None and isinstance(node, QolsysThermostat):
                                    node.update_thermostat(content_values)

                            # Update DoorLockContentProvider
                            case self.db.table_doorlock.uri:
                                self.db.table_doorlock.update(selection, selection_argument, content_values)
                                node_id = content_values.get("node_id", "")
                                node = self._controller.state.zwave_device(node_id)
                                if node is not None and isinstance(node, QolsysLock):
                                    node.update_lock(content_values)

                            # Update ZwaveContentProvider
                            case self.db.table_zwave_node.uri:
                                self.db.table_zwave_node.update(selection, selection_argument, content_values)
                                node_id = content_values.get("node_id", "")
                                node = self._controller.state.zwave_device(node_id)
                                if node is not None:
                                    node.update_base(content_values)

                            # Update Z-Wave History Content Provier
                            case self.db.table_zwave_history.uri:
                                self.db.table_zwave_history.update(selection, selection_argument, content_values)

                            # Update AutomationDeviceContentProvider
                            case self.db.table_automation.uri:
                                self.db.table_automation.update(selection, selection_argument, content_values)

                            # Update Alarmed Sensor Content Provider
                            case self.db.table_alarmedsensor.uri:
                                self.db.table_alarmedsensor.update(selection, selection_argument, content_values)

                            # Update IQ Remote Settings Content Provider
                            case self.db.table_iqremotesettings.uri:
                                self.db.table_iqremotesettings.update(selection, selection_argument, content_values)

                            # Update Scene Content Provider
                            case self.db.table_scene.uri:
                                self.db.table_scene.update(selection, selection_argument, content_values)
                                scene_id = content_values.get("scene_id", "")
                                scene = self._controller.state.scene(scene_id)
                                if scene is not None and isinstance(scene, QolsysScene):
                                    scene.update(content_values)

                            # Update Trouble Conditions
                            case self.db.table_trouble_conditions.uri:
                                self.db.table_trouble_conditions.update(selection, selection_argument, content_values)

                            # Update EU_EVENT:
                            case self.db.table_eu_event.uri:
                                self.db.table_eu_event.update(selection, selection_argument, content_values)

                            # Update PowerG Device
                            case self.db.table_powerg_device.uri:
                                self.db.table_powerg_device.update(selection, selection_argument, content_values)
                                short_id = content_values.get("shortID", "")
                                zone = self._controller.state.zone_from_short_id(short_id)
                                if zone is not None:
                                    zone.update_powerg(content_values)

                            # Update Weather
                            case self.db.table_weather.uri:
                                self.db.table_weather.update(selection, selection_argument, content_values)
                                self._controller.state.sync_weather_data(self.get_weather_from_db())

                            # Update Zwave Association Group
                            case self.db.table_zwave_association_goup.uri:
                                self.db.table_zwave_association_goup.update(selection, selection_argument, content_values)

                            # Country Locale
                            case self.db.table_country_locale.uri:
                                self.db.table_country_locale.update(selection, selection_argument, content_values)

                            # Virtual device
                            case self.db.table_virtual_device.uri:
                                self.db.table_virtual_device.update(selection, selection_argument, content_values)
                                adc_id = content_values.get("device_id", "")
                                adc_device = self._controller.state.adc_device(adc_id)
                                if adc_device is not None:
                                    adc_device.update_adc_device(content_values)

                            case _:
                                LOGGER.debug("iq2meid updating unknow uri:%s", uri)
                                LOGGER.debug(data)

                    case "delete":
                        selection = data.get("selection")
                        selection_argument = data.get("selectionArgs")

                        match uri:
                            case self.db.table_sensor.uri:
                                self.db.table_sensor.delete(selection, selection_argument)
                                self._controller.state.sync_zones_data(self.get_zones_from_db())

                            case self.db.table_iqremotesettings.uri:
                                self.db.table_iqremotesettings.delete(selection, selection_argument)

                            case self.db.table_state.uri:
                                self.db.table_state.delete(selection, selection_argument)

                            case self.db.table_master_slave.uri:
                                self.db.table_master_slave.delete(selection, selection_argument)

                            case self.db.table_qolsyssettings.uri:
                                self.db.table_qolsyssettings.delete(selection, selection_argument)

                            case self.db.table_alarmedsensor.uri:
                                self.db.table_alarmedsensor.delete(selection, selection_argument)
                                self._controller.state.sync_partitions_data(self.get_partitions_from_db())

                            case self.db.table_history.uri:
                                self.db.table_history.delete(selection, selection_argument)

                            case self.db.table_zwave_history.uri:
                                self.db.table_zwave_history.delete(selection, selection_argument)

                            case self.db.table_doorlock.uri:
                                self.db.table_doorlock.delete(selection, selection_argument)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            case self.db.table_dimmer.uri:
                                self.db.table_dimmer.delete(selection, selection_argument)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            case self.db.table_thermostat.uri:
                                self.db.table_thermostat.delete(selection, selection_argument)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            case self.db.table_zwave_node.uri:
                                self.db.table_zwave_node.delete(selection, selection_argument)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            case self.db.table_automation.uri:
                                self.db.table_automation.delete(selection, selection_argument)

                            case self.db.table_partition.uri:
                                self.db.table_partition.delete(selection, selection_argument)
                                self._controller.state.sync_partitions_data(self.get_partitions_from_db())

                            case self.db.table_user.uri:
                                self.db.table_user.delete(selection, selection_argument)

                            case self.db.table_dashboard_msgs.uri:
                                self.db.table_dashboard_msgs.delete(selection, selection_argument)

                            case self.db.table_eu_event.uri:
                                self.db.table_eu_event.delete(selection, selection_argument)

                            case self.db.table_powerg_device.uri:
                                self.db.table_powerg_device.delete(selection, selection_argument)

                            case self.db.table_weather.uri:
                                self.db.table_weather.delete(selection, selection_argument)
                                self._controller.state.sync_weather_data(self.get_weather_from_db())

                            case self.db.table_zwave_association_goup.uri:
                                self.db.table_zwave_association_goup.delete(selection, selection_argument)

                            case self.db.table_virtual_device.uri:
                                self.db.table_virtual_device.delete(selection, selection_argument)
                                self._controller.state.sync_adc_devices_data(self.get_adc_devices_from_db())

                            case _:
                                LOGGER.debug("iq2meid deleting unknown uri:%s", uri)
                                LOGGER.debug(data)

                    case "insert":
                        content_values = data.get("contentValues", {})

                        match uri:
                            # Inser State Content Provider
                            case self.db.table_state.uri:
                                self.db.table_state.insert(data=content_values)

                                name = content_values.get("name", "")
                                new_value = content_values.get("value", "")
                                if name in self.state_partition:
                                    partition_id = content_values.get("partition_id", "")
                                    partition = self._controller.state.partition(partition_id)
                                    if partition is not None:
                                        match name:
                                            case "ALARM_STATE":
                                                partition.alarm_state = PartitionAlarmState(new_value)

                            # Inser Partition Content Provider
                            case self.db.table_partition.uri:
                                self.db.table_partition.insert(data=content_values)
                                self._controller.state.sync_partitions_data(self.get_partitions_from_db())

                            # Insert Settings Content Provider
                            case self.db.table_qolsyssettings.uri:
                                self.db.table_qolsyssettings.insert(data=content_values)

                                # Update Partition setting - Send notification if setting has changed
                                name = content_values.get("name", "")
                                new_value = content_values.get("value", "")
                                if name in self.settings_partition:
                                    partition_id = content_values.get("partition_id", "")
                                    partition = self._controller.state.partition(partition_id)
                                    if partition is not None:
                                        match name:
                                            case "SYSTEM_STATUS":
                                                partition.system_status = PartitionSystemStatus(new_value)
                                            case "SYSTEM_STATUS_CHANGED_TIME":
                                                partition.system_status_changed_time = new_value
                                            case "EXIT_SOUNDS":
                                                partition.exit_sounds = new_value
                                            case "ENTRY_DELAYS":
                                                partition.entry_delays = new_value

                            # UserContentProvider
                            case self.db.table_user.uri:
                                self.db.table_user.insert(data=content_values)
                                # No action needed

                            # MasterSlave Content Provider
                            case self.db.table_master_slave.uri:
                                self.db.table_master_slave.insert(data=content_values)
                                # No action needed

                            # Automation Content Provider
                            case self.db.table_automation.uri:
                                self.db.table_automation.insert(content_values)
                                # No action needed

                            # Sensor Content Provider
                            case self.db.table_sensor.uri:
                                self.db.table_sensor.insert(data=content_values)
                                self._controller.state.sync_zones_data(self.get_zones_from_db())

                            # Door Lock Content Provider
                            case self.db.table_doorlock.uri:
                                self.db.table_doorlock.insert(data=content_values)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            # Dimmer Content Provider
                            case self.db.table_dimmer.uri:
                                self.db.table_dimmer.insert(data=content_values)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            # Thermostat Content Provider
                            case self.db.table_thermostat.uri:
                                self.db.table_thermostat.insert(data=content_values)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            # ZWave Node Content Provider
                            case self.db.table_zwave_node.uri:
                                self.db.table_zwave_node.insert(data=content_values)
                                self._controller.state.sync_zwave_devices_data(self.get_zwave_devices_from_db())

                            # HistoryContentProvider
                            case self.db.table_history.uri:
                                self.db.table_history.insert(data=content_values)

                            # AlarmedSensorProvider
                            case self.db.table_alarmedsensor.uri:
                                partition_id = content_values.get("partition_id", "")
                                self.db.table_alarmedsensor.insert(data=content_values)

                                partition = self._controller.state.partition(partition_id)
                                if partition is not None:
                                    partition.append_alarm_type([PartitionAlarmType(content_values.get("sgroup", ""))])

                            # IQRemoteSettingsProvider
                            case self.db.table_iqremotesettings.uri:
                                self.db.table_iqremotesettings.insert(data=content_values)

                            # HeatMapContentProvider
                            case self.db.table_heat_map.uri:
                                self.db.table_heat_map.insert(data=content_values)

                            # ZDeviceHistoryContentProvider
                            case self.db.table_zwave_history.uri:
                                self.db.table_zwave_history.insert(data=content_values)

                            # Dashboard Message Content Provider
                            case self.db.table_dashboard_msgs.uri:
                                self.db.table_dashboard_msgs.insert(data=content_values)

                            # EU_EVENT
                            case self.db.table_eu_event.uri:
                                self.db.table_eu_event.insert(data=content_values)

                            # PowerG Device
                            case self.db.table_powerg_device.uri:
                                self.db.table_powerg_device.insert(data=content_values)

                            # Weather
                            case self.db.table_weather.uri:
                                self.db.table_weather.insert(data=content_values)
                                self._controller.state.sync_weather_data(self.get_weather_from_db())

                            # ZWave Association Group
                            case self.db.table_zwave_association_goup.uri:
                                self.db.table_zwave_association_goup.insert(data=content_values)

                            # Virtual Device
                            case self.db.table_virtual_device.uri:
                                self.db.table_virtual_device.insert(data=content_values)
                                self._controller.state.sync_adc_devices_data(self.get_adc_devices_from_db())

                            case _:
                                LOGGER.debug("iq2meid inserting unknow uri:%s", uri)
                                LOGGER.debug(data)

                    case _:
                        LOGGER.debug("iq2meid - Unknow dboperation: %s", dbOperation)
                        LOGGER.debug(data)
            case _:
                LOGGER.debug("iq2meid - Unknow event: %s", eventName)
                LOGGER.debug(data)

    def check_user(self, user_code: str) -> int:
        for user in self._users:
            if user.user_code == user_code:
                return user.id

        # No valid user code found
        return -1

    def get_adc_devices_from_db(self) -> list[QolsysAdcDevice]:
        adc_devices: list[QolsysAdcDevice] = []
        devices_list = self.db.get_adc_devices()

        for device in devices_list:
            adc_devices.append(QolsysAdcDevice(device))

        return adc_devices

    def get_zwave_devices_from_db(self) -> list[QolsysZWaveDevice]:
        devices: list[QolsysZWaveDevice] = []
        devices_list = self.db.get_zwave_devices()
        dimmers_list = self.db.get_dimmers()
        thermostats_list = self.db.get_thermostats()
        locks_list = self.db.get_locks()

        for device in devices_list:
            device_added = False
            zwave_node_id = device.get("node_id", "")

            # Check if z-wave device is an Energy Clamp
            if device.get("node_type", "") == "Energy Clamp":
                qolsys_power = QolsysPower(zwave_dict=device)
                devices.append(qolsys_power)
                device_added = True

            # Check if z-wave device is a thermometer
            if device.get("node_type", "") == "Thermometer":
                qolsys_thermometer = QolsysThermometer(device)
                devices.append(qolsys_thermometer)
                device_added = True

            # Check if z-wave device is a Dimmer
            for d in dimmers_list:
                dimmer_node_id = d.get("node_id", "")

                # Found a Dimmer
                if zwave_node_id == dimmer_node_id:
                    qolsys_dimmer = QolsysDimmer(d, device)
                    devices.append(qolsys_dimmer)
                    device_added = True
                    break

            # Check is z-wave devie is a Thermostat
            for thermostat in thermostats_list:
                thermostat_node_id = thermostat.get("node_id", "")

                # Found a Thermostat
                if zwave_node_id == thermostat_node_id:
                    qolsys_thermostat = QolsysThermostat(thermostat, device)
                    devices.append(qolsys_thermostat)
                    device_added = True
                    break

            # Check is z-wave device is a Lock
            for lock in locks_list:
                lock_node_id = lock.get("node_id", "")

                # Found a Lock
                if zwave_node_id == lock_node_id:
                    qolsys_lock = QolsysLock(lock, device)
                    devices.append(qolsys_lock)
                    device_added = True
                    break

            # Found a Smart Outlet

            # Found Garage Door Openner

            # No Specific z-wave device found, add a generic z-wave device
            if not device_added:
                qolsys_generic = QolsysGeneric(device)
                devices.append(qolsys_generic)

        return devices

    def get_scenes_from_db(self) -> list[QolsysScene]:
        scenes = []
        scenes_list: list[dict[str, str]] = self.db.get_scenes()

        # Create scenes array
        for scene_info in scenes_list:
            scenes.append(QolsysScene(scene_info))

        return scenes

    def get_weather_from_db(self) -> QolsysWeather:
        weather = QolsysWeather()
        forecast_dic_list: list[dict[str, str]] = self.db.get_weather()

        forecast_obj_list = []
        for forecast in forecast_dic_list:
            forecast_obj_list.append(QolsysForecast(forecast))

        # Create weather array
        weather.update(forecast_obj_list)

        return weather

    def get_zones_from_db(self) -> list[QolsysZone]:
        zones = []
        zones_list: list[dict[str, str]] = self.db.get_zones()

        # Create sensors array
        for zone_info in zones_list:
            new_zone = QolsysZone(zone_info, self._controller.settings)

            if new_zone.current_capability == "POWERG":
                LOGGER.debug("Loading PowerG device info for zone %s", new_zone.zone_id)
                powerg_dict = self.db.get_powerg(short_id=new_zone.shortID)
                LOGGER.debug("PowerG device info: %s", powerg_dict)
                if powerg_dict is not None:
                    new_zone.update_powerg(powerg_dict)

            zones.append(new_zone)

        return zones

    def get_partitions_from_db(self) -> list[QolsysPartition]:
        partitions = []
        partition_list: list[dict[str, str]] = self.db.get_partitions()

        # Create partitions array
        for partition_dict in partition_list:
            partition_id = partition_dict["partition_id"]

            settings_dict = {
                "SYSTEM_STATUS": self.db.get_setting_partition("SYSTEM_STATUS", partition_id) or "UNKNOWN",
                "SYSTEM_STATUS_CHANGED_TIME": self.db.get_setting_partition("SYSTEM_STATUS_CHANGED_TIME", partition_id) or "",
                "EXIT_SOUNDS": self.db.get_setting_partition("EXIT_SOUNDS", partition_id) or "",
                "ENTRY_DELAYS": self.db.get_setting_partition("ENTRY_DELAYS", partition_id) or "",
            }

            alarm_type = []
            for alarm in self.db.get_alarm_type(partition_id):
                alarm_type.append(PartitionAlarmType(alarm))

            alarm_state = PartitionAlarmState(self.db.get_state_partition("ALARM_STATE", partition_id) or "UNKNOWN")

            partition = QolsysPartition(partition_dict, settings_dict, alarm_state, alarm_type)
            partitions.append(partition)

        return partitions

    def dump(self) -> None:
        LOGGER.debug("*** Qolsys Panel Information ***")
        LOGGER.debug("Android Version: %s", self.ANDROID_VERSION)
        LOGGER.debug("Hardware Version: %s", self.HARDWARE_VERSION)
        LOGGER.debug("MAC Address: %s", self.MAC_ADDRESS)
        LOGGER.debug("Unique ID: %s", self.unique_id)
        LOGGER.debug("Panel Tamper State: %s", self.PANEL_TAMPER_STATE)
        LOGGER.debug("AC Status: %s", self.AC_STATUS)
        LOGGER.debug("Battery Status: %s", self.BATTERY_STATUS)
        LOGGER.debug("GSM Connection Status: %s", self.GSM_CONNECTION_STATUS)
        LOGGER.debug("GSM Signal Strength: %s", self.GSM_SIGNAL_STRENGTH)
        LOGGER.debug("Fail To Communicate: %s", self.FAIL_TO_COMMUNICATE)
        # LOGGER.debug("System Time: %s",datetime.fromtimestamp(int(self.SYSTEM_TIME)/1000))
        LOGGER.debug("Country: %s", self.COUNTRY)
        LOGGER.debug("Language: %s", self.LANGUAGE)
        LOGGER.debug("Temp Format: %s", self.TEMPFORMAT)
        LOGGER.debug("Z-Wave Firmware Version: %s", self.ZWAVE_FIRM_WARE_VERSION)
        LOGGER.debug("Z-Wave Card Present: %s", self.ZWAVE_CARD)
        LOGGER.debug("Z-Wave Controller Enabled: %s", self.ZWAVE_CONTROLLER)
        LOGGER.debug("Partitons Enabled: %s", self.PARTITIONS)
        LOGGER.debug("Control4 Enabled: %s", self.CONTROL_4)
        LOGGER.debug("Six Digit User Code Enabled: %s", self.SIX_DIGIT_USER_CODE)
        LOGGER.debug("Secure Arming: %s", self.SECURE_ARMING)
        LOGGER.debug("Auto-Stay: %s", self.AUTO_STAY)
        LOGGER.debug("Auto-Bypass: %s", self.AUTO_BYPASS)
        LOGGER.debug("Auto-Arm-Stay: %s", self.AUTO_ARM_STAY)
        LOGGER.debug("Auto-Exit-Extension: %s", self.AUTO_EXIT_EXTENSION)
        LOGGER.debug("Final-Exit-Door-Arming: %s", self.FINAL_EXIT_DOOR_ARMING)
        LOGGER.debug("No-Arm-Low-Battery: %s", self.NO_ARM_LOW_BATTERY)
        LOGGER.debug("Normal Entry Delay: %s", self.TIMER_NORMAL_ENTRY_DELAY)
        LOGGER.debug("Normal Exit Delay: %s", self.TIMER_NORMAL_EXIT_DELAY)
        LOGGER.debug("Long Entry Delay: %s", self.TIMER_LONG_ENTRY_DELAY)
        LOGGER.debug("Long Exit Delay: %s", self.TIMER_LONG_EXIT_DELAY)
        LOGGER.debug("Auxiliary Panic Enabled: %s", self.AUXILIARY_PANIC_ENABLED)
        LOGGER.debug("Fire Panic Enabled: %s", self.FIRE_PANIC_ENABLED)
        LOGGER.debug("Police Panic Enabled: %s", self.POLICE_PANIC_ENABLED)
        LOGGER.debug("Night Mode Settings: %s", self.NIGHTMODE_SETTINGS)
        LOGGER.debug("Night Mode Settings Stage: %s", self.NIGHT_SETTINGS_STATE)
        LOGGER.debug("Show Security Sensors: %s", self.SHOW_SECURITY_SENSORS)

        LOGGER.debug("Users list:")
        for user in self._users:
            LOGGER.debug("User: %s", user.id)

        LOGGER.debug("*** Plugin Information ***")
        LOGGER.debug("Motion Delay Enabled: %s", self._controller.settings.motion_sensor_delay)
        LOGGER.debug("Motion Delay Value: %s", self._controller.settings.motion_sensor_delay_sec)



import logging

from .zwave_device import QolsysZWaveDevice

LOGGER = logging.getLogger(__name__)


class QolsysGarageDoor(QolsysZWaveDevice):
    def __init__(self) -> None:
        pass

    def update_raw(self, command: int, command_status: int, command_type: int, payload: list[int]) -> None:
        LOGGER.debug(
            "Raw Update (node%s) - command:%s status:%s type:%s payload:%s",
            self.node_id,
            command,
            command_status,
            command_type,
            payload,
        )



import logging

LOGGER = logging.getLogger(__name__)


class QolsysUser:
    def __init__(self) -> None:
        self._id: int = 0
        self._user_code = ""

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, value: int) -> None:
        self._id = value

    @property
    def user_code(self) -> str:
        return self._user_code

    @user_code.setter
    def user_code(self, value: str) -> None:
        self._user_code = value



#!/usr/bin/env python3

import asyncio
import logging
import os
import sys

from qolsys_controller.controller import QolsysController
from qolsys_controller.errors import QolsysMqttError, QolsysSqlError, QolsysSslError

logging.basicConfig(level=logging.DEBUG, format="%(levelname)s - %(module)s: %(message)s")
LOGGER = logging.getLogger(__name__)


async def main() -> None:  # noqa: D103
    remote = QolsysController()
    remote.settings.config_directory = "./config/"
    remote.settings.panel_ip = "****"
    remote.settings.plugin_ip = "****"
    remote.settings.random_mac = ""  # Example: F2:16:3E:33:ED:20

    # Additionnal remote plugin config
    remote.settings.check_user_code_on_disarm = False  # Check user code in user.conf file
    remote.settings.log_mqtt_mesages = False  # Enable for MQTT debug purposes
    remote.settings.auto_discover_pki = True

    # Configure remote plugin
    if not await remote.config(start_pairing=True):
        LOGGER.debug("Error Configuring remote plugin")
        return

    try:
        await remote.start_operation()

    except QolsysMqttError:
        LOGGER.debug("QolsysMqttError")

    except QolsysSslError:
        LOGGER.debug("QolsysSslError")

    except QolsysSqlError:
        LOGGER.debug("QolsysSqlError")

    if not remote.connected:
        LOGGER.error("Panel not ready for operation")
        return

    LOGGER.debug("Qolsys Panel Ready for operation")

    await asyncio.sleep(5)
    # await remote.command_zwave_switch_multi_level(node_id="6",level=99)
    # await remote.command_zwave_switch_binary_set(node_id="8", status=True)
    # await remote.plugin.command_arm(
    #   partition_id='0',
    #   arming_type=PartitionArmingType:ARM_NIGHT,
    #   user_code="1111",
    #   exit_sounds=False,
    #   instant_arm=True)
    # await remote.command_disarm(partition_id="0", user_code="1111")
    # await remote.command_panel_execute_scene(scene_id="3")
    # await remote.command_zwave_doorlock_set(node_id="7",locked=True)
    # await remote.stop_operation()

    # Use an asyncio.Event to keep the program running efficiently
    stop_event = asyncio.Event()
    await stop_event.wait()


# Change to the "Selector" event loop if platform is Windows
if sys.platform.lower() == "win32" or os.name.lower() == "nt":
    from asyncio import (  # type: ignore[attr-defined]
        WindowsSelectorEventLoopPolicy,
        set_event_loop_policy,
    )

    set_event_loop_policy(WindowsSelectorEventLoopPolicy())

asyncio.run(main())



import json
import logging
from typing import TYPE_CHECKING

from .adc_service import QolsysAdcService
from .adc_service_garagedoor import QolsysAdcGarageDoorService
from .enum_adc import vdFuncLocalControl, vdFuncName, vdFuncState, vdFuncType
from .observable import QolsysObservable

if TYPE_CHECKING:
    from .adc_service import QolsysAdcService

LOGGER = logging.getLogger(__name__)


class QolsysAdcDevice(QolsysObservable):
    def __init__(self, adc_dict: dict[str, str]) -> None:
        super().__init__()

        self._services: list[QolsysAdcService] = []

        self._id: str = adc_dict.get("_id", "")
        self._partition_id: str = adc_dict.get("partition_id", "")
        self._device_id: str = adc_dict.get("device_id", "")
        self._name: str = adc_dict.get("name", "")
        self._type: str = adc_dict.get("type", "")
        self._create_time: str = adc_dict.get("create_time", "")
        self._created_by: str = adc_dict.get("created_by", "")
        self._update_time: str = adc_dict.get("update_time", "")
        self._updated_by: str = adc_dict.get("updated_by", "")
        self._device_zone_list: str = adc_dict.get("device_zone_list", "")
        self._func_list = ""
        self.func_list = adc_dict.get("func_list", "")

    def update_adc_device(self, data: dict[str, str]) -> None:
        # Check if we are updating same device_id
        device_id_update = data.get("device_id", "")
        if device_id_update != self._device_id:
            LOGGER.error(
                "Updating ADC%s (%s) with ADC%s (different device_id)",
                self._device_id,
                self._name,
                device_id_update,
            )
            return

        self.start_batch_update()

        if "partition_id" in data:
            self.partition_id = data.get("partition_id", "")

        if "name" in data:
            self.name = data.get("name", "")

        if "type" in data:
            self.type = data.get("type", "")

        self.end_batch_update()

        if "func_list" in data:
            self.func_list = data.get("func_list", "")

    def get_adc_service(self, id: int) -> QolsysAdcService | None:
        for service in self._services:
            if service.id == id:
                return service
        return None

    def add_adc_service(
        self,
        id: int,
        local_control: vdFuncLocalControl,
        func_name: vdFuncName,
        func_type: vdFuncType,
        func_state: vdFuncState,
        timestamp: str,
    ) -> None:
        # Garage Door
        if func_name == vdFuncName.OPEN_CLOSE and func_type == vdFuncType.BINARY_ACTUATOR:
            LOGGER.debug("ADC%s (%s) - Adding garage door service", self.device_id, self.name)
            self._services.append(
                QolsysAdcGarageDoorService(self, id, func_name, local_control, func_type, func_state, timestamp)
            )
            self.notify()
            return

        # Add generic service if other services have beed identified
        LOGGER.debug("ADC%s (%s) - Adding generic service", self.device_id, self.name)
        self._services.append(QolsysAdcService(self, id, func_name, local_control, func_type, func_state, timestamp))
        self.notify()
        return

    # -----------------------------
    # properties + setters
    # -----------------------------

    @property
    def device_id(self) -> str:
        return self._device_id

    @device_id.setter
    def device_id(self, value: str) -> None:
        self._device_id = value

    @property
    def services(self) -> list[QolsysAdcService]:
        return self._services

    @property
    def partition_id(self) -> str:
        return self._partition_id

    @partition_id.setter
    def partition_id(self, value: str) -> None:
        if self._partition_id != value:
            LOGGER.debug("ADC%s (%s) - partition_id: %s", self.device_id, self.name, value)
            self._partition_id = value
            self.notify()

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        if self._name != value:
            LOGGER.debug("ADC%s (%s) - name: %s", self.device_id, self.name, value)
            self._name = value
            self.notify()

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, value: str) -> None:
        if self._type != value:
            LOGGER.debug("ADC%s (%s) - type: %s", self.device_id, self.name, value)
            self._type = value
            self.notify()

    @property
    def func_list(self) -> str:
        return self._func_list

    @func_list.setter
    def func_list(self, value: str) -> None:
        if self._func_list != value:
            LOGGER.debug("ADC%s (%s) - func_list: %s", self.device_id, self.name, value)
            self._func_list = value

            try:
                json_func_list = json.loads(self._func_list)
                new_service_id: list[int] = []
                self.start_batch_update()

                for function in json_func_list:
                    try:
                        id = function.get("vdFuncId")
                        local_control = vdFuncLocalControl(function.get("vdFuncLocalControl"))
                        func_name = vdFuncName(function.get("vdFuncName"))
                        func_type = vdFuncType(function.get("vdFuncType"))
                        func_state = vdFuncState(function.get("vdFuncState"))
                        timestamp = function.get("vdFuncBackendTimestamp")
                        new_service_id.append(id)

                        service = self.get_adc_service(id)
                        if service is not None:
                            service.update_adc_service(func_name, local_control, func_type, func_state, timestamp)

                        if service is None:
                            self.add_adc_service(id, local_control, func_name, func_type, func_state, timestamp)

                    except ValueError as e:
                        LOGGER.error("Error converting value:", e)
                        continue

                # Check if service have been removed
                for service in self._services:
                    if service.id not in new_service_id:
                        self._services.remove(service)
                        self.notify()

                self.end_batch_update()

            except json.JSONDecodeError as e:
                LOGGER.error("ADC%s - Error parsing JSON:", self.device_id, e)

    def to_dict_adc(self) -> dict[str, str]:
        return {
            "_id": self._id,
            "partition_id": self.partition_id,
            "device_id": self.device_id,
            "name": self.name,
            "type": self.type,
            "func_list": self.func_list,
            "create_time": self._create_time,
            "created_by": self._created_by,
            "update_time": self._update_time,
            "updated_by": self._updated_by,
            "device_zone_list": self._device_zone_list,
        }



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .controller import QolsysController


class QolsysSettings:
    def __init__(self, controller: QolsysController) -> None:
        self._controller = controller

        # Plugin
        self._plugin_ip: str = ""
        self._random_mac: str = ""
        self._panel_mac: str = ""
        self._panel_ip: str = ""

        # Path
        self._config_directory: Path = Path()
        self._pki_directory: Path = Path()
        self._media_directory: Path = Path()
        self._users_file_path: Path = Path()

        # Pki
        self._key_size: int = 2048
        self._auto_discover_pki: bool = True

        # MQTT
        self._mqtt_timeout: int = 30
        self._mqtt_ping: int = 600
        self._mqtt_qos: int = 0
        self._mqtt_remote_client_id: str = ""
        self._log_mqtt_message: bool = False

        # Operation
        self._motion_sensor_delay: bool = True
        self._motion_sensor_delay_sec: int = 310
        self._check_user_code_on_arm: bool = False
        self._check_user_code_on_disarm: bool = True

    def check_panel_ip(self) -> bool:
        if self._panel_ip == "":
            LOGGER.debug("Invalid Panel IP:  %s", self._panel_ip)
            return False

        LOGGER.debug("Found Panel IP: %s", self._panel_ip)
        return True

    def check_plugin_ip(self) -> bool:
        if self._plugin_ip == "":
            LOGGER.debug("Invalid Plugin IP:  %s", self._plugin_ip)
            return False

        LOGGER.debug("Found Plugin IP: %s", self._plugin_ip)
        return True

    # -----------------------------
    # properties + setters
    # -----------------------------

    @property
    def random_mac(self) -> str:
        return self._random_mac

    @random_mac.setter
    def random_mac(self, random_mac: str) -> None:
        self._random_mac = random_mac.lower()

    @property
    def plugin_ip(self) -> str:
        return self._plugin_ip

    @plugin_ip.setter
    def plugin_ip(self, plugin_ip: str) -> None:
        self._plugin_ip = plugin_ip

    @property
    def panel_mac(self) -> str:
        return self._panel_mac

    @panel_mac.setter
    def panel_mac(self, panel_mac: str) -> None:
        self._panel_mac = panel_mac

    @property
    def panel_ip(self) -> str:
        return self._panel_ip

    @panel_ip.setter
    def panel_ip(self, panel_ip: str) -> None:
        self._panel_ip = panel_ip

    @property
    def log_mqtt_mesages(self) -> bool:
        return self._log_mqtt_messages

    @log_mqtt_mesages.setter
    def log_mqtt_mesages(self, log_mqtt_mesages: bool) -> None:
        self._log_mqtt_messages = log_mqtt_mesages

    @property
    def check_user_code_on_disarm(self) -> bool:
        return self._check_user_code_on_disarm

    @check_user_code_on_disarm.setter
    def check_user_code_on_disarm(self, check_user_code_on_disarm: bool) -> None:
        self._check_user_code_on_disarm = check_user_code_on_disarm

    @property
    def check_user_code_on_arm(self) -> bool:
        return self._check_user_code_on_arm

    @check_user_code_on_arm.setter
    def check_user_code_on_arm(self, check_user_code_on_arm: bool) -> None:
        self._check_user_code_on_arm = check_user_code_on_arm

    @property
    def auto_discover_pki(self) -> bool:
        return self._auto_discover_pki

    @auto_discover_pki.setter
    def auto_discover_pki(self, value: bool) -> None:
        self._auto_discover_pki = value

    @property
    def mqtt_timeout(self) -> int:
        return self._mqtt_timeout

    @mqtt_timeout.setter
    def mqtt_timeout(self, value: int) -> None:
        self._mqtt_timeout = value

    @property
    def mqtt_ping(self) -> int:
        return self._mqtt_ping

    @mqtt_ping.setter
    def mqtt_ping(self, value: int) -> None:
        self._mqtt_ping = value

    @property
    def motion_sensor_delay(self) -> bool:
        return self._motion_sensor_delay

    @motion_sensor_delay.setter
    def motion_sensor_delay(self, value: bool) -> None:
        self._motion_sensor_delay = value

    @property
    def motion_sensor_delay_sec(self) -> int:
        return self._motion_sensor_delay_sec

    @motion_sensor_delay_sec.setter
    def motion_sensor_delay_sec(self, value: int) -> None:
        self._motion_sensor_delay_sec = value

    @property
    def config_directory(self) -> Path:
        return self._config_directory

    @config_directory.setter
    def config_directory(self, config_directory: str) -> None:
        self._config_directory = Path(config_directory)
        self._pki_directory = self._config_directory.joinpath("pki")
        self._media_directory = self._config_directory.joinpath("media")
        self._users_file_path = self._config_directory.joinpath("users.conf")

    @property
    def pki_directory(self) -> Path:
        return self._pki_directory

    @property
    def users_file_path(self) -> Path:
        return self._users_file_path

    @property
    def key_size(self) -> int:
        return self._key_size

    @property
    def mqtt_qos(self) -> int:
        return self._mqtt_qos

    @property
    def mqtt_remote_client_id(self) -> str:
        return self._mqtt_remote_client_id

    @mqtt_remote_client_id.setter
    def mqtt_remote_client_id(self, client_id: str) -> None:
        self._mqtt_remote_client_id = client_id

    def check_config_directory(self, create: bool = True) -> bool:  # noqa: PLR0911
        if not self.config_directory.is_dir():
            if not create:
                LOGGER.debug("config_directory not found:  %s", self.config_directory)
                return False

            # Create config directory if not found
            LOGGER.debug("Creating config_directory: %s", self.config_directory)
            try:
                self.config_directory.mkdir(parents=True)
            except PermissionError:
                LOGGER.exception("Permission denied: Unable to create: %s", self.config_directory)
                return False
            except Exception:
                LOGGER.exception("Error creating config_directory: %s", self.config_directory)
                return False

        LOGGER.debug("Using config_directory: %s", self.config_directory.resolve())

        # Create pki directory if not found
        if not self.pki_directory.is_dir():
            LOGGER.debug("Creating pki_directory: %s", self.pki_directory.resolve())
            try:
                self.pki_directory.mkdir(parents=True)
            except PermissionError:
                LOGGER.exception("Permission denied: Unable to create: %s", self.pki_directory.resolve())
                return False
            except Exception:
                LOGGER.exception("Error creating pki_directory: %s", self.pki_directory.resolve())
                return False

        LOGGER.debug("Using pki_directory: %s", self.pki_directory.resolve())

        # Create media directory if not found
        if not self._media_directory.is_dir():
            LOGGER.debug("Creating media_directory: %s", self._media_directory.resolve())
            try:
                self._media_directory.mkdir(parents=True)
            except PermissionError:
                LOGGER.exception("Permission denied: Unable to create: %s", self._media_directory.resolve())
                return False
            except Exception:
                LOGGER.exception("Error creating media_directory: %s", self._media_directory.resolve())
                return False

        LOGGER.debug("Using media_directory: %s", self._media_directory.resolve())

        return True



import logging  # noqa: INP001
import sqlite3

from qolsys_controller.errors import QolsysSqlError

LOGGER = logging.getLogger(__name__)


class QolsysTable:
    def __init__(self, db: sqlite3.Connection, cursor: sqlite3.Cursor) -> None:
        self._db: sqlite3.Connection = db
        self._cursor: sqlite3.Cursor = cursor
        self._uri: str = ""
        self._table: str = ""
        self._columns: list[str] = []
        self._abort_on_error: bool = False
        self._implemented: bool = False

    @property
    def uri(self) -> str:
        return self._uri

    @property
    def table(self) -> str:
        return self._table

    def _create_table(self) -> None:
        if not self._columns:
            msg = "The column list must not be empty."
            raise ValueError(msg)

        primary_key = self._columns[0]
        other_columns = self._columns[1:]

        column_defs = [f"{primary_key} TEXT PRIMARY KEY"]
        column_defs += [f"{col} TEXT" for col in other_columns]

        try:
            query: str = f"CREATE TABLE {self._table} ({', '.join(column_defs)})"
            self._cursor.execute(query)
            self._db.commit()

        except sqlite3.Error as err:
            error = QolsysSqlError(
                {
                    "table": self.table,
                    "query": query,
                    "columns": self._columns,
                }
            )

            if self._abort_on_error:
                raise error from err

    def clear(self) -> None:
        try:
            query = f"DELETE from {self.table}"
            self._cursor.execute(query)
            self._db.commit()

        except sqlite3.Error as err:
            error = QolsysSqlError(
                {
                    "table": self.table,
                    "query": query,
                    "columns": self._columns,
                }
            )

            if self._abort_on_error:
                raise error from err

    def insert(self, data: dict[str, str]) -> None:
        try:
            if not self._implemented and data is not None:
                LOGGER.warning("New Table format: %s", self.uri)
                LOGGER.warning("Table: %s", self.table)
                LOGGER.warning(data)
                LOGGER.warning("Please Report")
                return

            # Select only known columns
            full_data = {col: data.get(col, "") for col in self._columns}

            new_columns = []
            for key in data:
                if key not in self._columns:
                    new_columns.append(key)

            # Warn if new column found in iq2meid database
            if new_columns:
                LOGGER.warning("New column found in iq2meid database")
                LOGGER.warning("Table: %s", self.table)
                LOGGER.warning("New Columns: %s", new_columns)
                LOGGER.warning("Please Report")

            col_str = ", ".join(full_data.keys())
            placeholder_str = ", ".join([f":{key}" for key in full_data])

            query = f"INSERT OR IGNORE INTO {self.table} ({col_str}) VALUES ({placeholder_str})"
            self._cursor.execute(query, full_data)
            self._db.commit()

        except sqlite3.Error as err:
            error = QolsysSqlError(
                {
                    "table": self.table,
                    "query": query,
                    "columns": self._columns,
                }
            )

            if self._abort_on_error:
                raise error from err

    def update(
        self, selection: str | None, selection_argument: list[str] | str | None, content_value: dict[str, str] | None
    ) -> None:
        # selection: 'zone_id=?, parition_id=?'
        # Firmware 4.4.1: selection_argument: '[3,1]'
        # Firmware 4.6.1: selection_argument: ['3','1']

        # Firmware 4.4.1: seletion_argument is sent as a string and needs to be converted to an array
        if isinstance(selection_argument, str):
            selection_argument = selection_argument.strip("[]")
            selection_argument = [item.strip() for item in selection_argument.split(",")]

        if selection_argument is None:
            selection_argument = []

        try:
            full_data = {}
            new_columns = []

            # Separate valid and unknown columns
            if content_value is not None:
                for key, value in content_value.items():
                    if key in self._columns:
                        full_data[key] = value
                    else:
                        new_columns.append(key)

            # Warn for unknown columns
            if new_columns:
                LOGGER.warning("New column found in iq2meid database")
                LOGGER.warning("Table: %s", self.table)
                LOGGER.warning("New Columns: %s", new_columns)
                LOGGER.warning("Please Report")

            set_clause = ", ".join([f"{key} = ?" for key in full_data])
            set_values = list(full_data.values())

            if selection:
                query = f"UPDATE {self.table} SET {set_clause} WHERE {selection}"
                params = set_values + selection_argument
            else:
                query = f"UPDATE {self.table} SET {set_clause}"
                params = set_values

            self._cursor.execute(query, params)
            self._db.commit()

        except sqlite3.Error as err:
            error = QolsysSqlError(
                {
                    "table": self.table,
                    "query": query,
                    "columns": self._columns,
                    "content_value": content_value,
                    "selection": selection,
                    "selection_argument": selection_argument,
                }
            )

            if self._abort_on_error:
                raise error from err

    def delete(self, selection: str | None, selection_argument: list[str] | str | None) -> None:
        # selection: 'zone_id=?, parition_id=?'
        # Firmware 4.4.1: selection_argument: '[3,1]'
        # Firmware 4.6.1: selection_argument: ['3','1']

        # Firmware 4.4.1: seletion_argument is sent as a string and needs to be converted to an array
        if type(selection_argument) is str:
            selection_argument = selection_argument.strip("[]")
            selection_argument = [item.strip() for item in selection_argument.split(",")]

        if selection_argument is None:
            selection_argument = []

        try:
            if selection:
                query = f"DELETE FROM {self.table} WHERE {selection}"

                if "?" in selection:
                    # Query expects parameters → must pass the list
                    self._cursor.execute(query, selection_argument)
                else:
                    # Query has no ? , do not pass arguments
                    self._cursor.execute(query)
            else:
                self.clear()

            self._db.commit()

        except sqlite3.Error as err:
            error = QolsysSqlError(
                {
                    "table": self.table,
                    "query": query,
                    "columns": self._columns,
                    "selection": selection,
                    "selection_argument": selection_argument,
                }
            )

            if self._abort_on_error:
                raise error from err
